#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""Utilities for building model-based closures."""

from __future__ import annotations

from collections.abc import Callable, Sequence
from itertools import chain, repeat
from types import NoneType
from typing import Any

from botorch.optim.closures.core import ForwardBackwardClosure
from botorch.utils.dispatcher import Dispatcher, type_bypassing_encoder
from gpytorch.mlls import (
    ExactMarginalLogLikelihood,
    MarginalLogLikelihood,
    SumMarginalLogLikelihood,
)
from torch import Tensor
from torch.utils.data import DataLoader

GetLossClosure = Dispatcher("get_loss_closure", encoder=type_bypassing_encoder)


def get_loss_closure(
    mll: MarginalLogLikelihood,
    data_loader: DataLoader | None = None,
    **kwargs: Any,
) -> Callable[[], Tensor]:
    r"""Public API for GetLossClosure dispatcher.

    This method, and the dispatcher that powers it, acts as a clearing house
    for factory functions that define how `mll` is evaluated.

    Users may specify custom evaluation routines by registering a factory function
    with GetLossClosure. These factories should be registered using the type signature

        `Type[MarginalLogLikeLihood], Type[Likelihood], Type[Model], Type[DataLoader]`.

    The final argument, Type[DataLoader], is optional. Evaluation routines that obtain
    training data from, e.g., `mll.model` should register this argument as `type(None)`.

    Args:
        mll: A MarginalLogLikelihood instance whose negative defines the loss.
        data_loader: An optional DataLoader instance for cases where training
            data is passed in rather than obtained from `mll.model`.

    Returns:
        A closure that takes zero positional arguments and returns the negated
        value of `mll`.
    """
    return GetLossClosure(
        mll, type(mll.likelihood), type(mll.model), data_loader, **kwargs
    )


def get_loss_closure_with_grads(
    mll: MarginalLogLikelihood,
    parameters: dict[str, Tensor],
    data_loader: DataLoader | None = None,
    **kwargs: Any,
) -> ForwardBackwardClosure:
    """
    Add a backward pass to a loss closure obtained by calling
    `get_loss_closure`, wrapping it in a ``ForwardBackwardClosure``.

    For further details, see `get_loss_closure`.

    Args:
        mll: A MarginalLogLikelihood instance whose negative defines the loss.
        parameters: A dictionary of tensors whose `grad` fields are to be returned.
        data_loader: An optional DataLoader instance for cases where training
            data is passed in rather than obtained from `mll.model`.
        kwargs: Keyword arguments passed to `get_loss_closure`.

    Returns:
        A closure that takes zero positional arguments and returns the reduced and
        negated value of `mll` along with the gradients of `parameters`.
    """
    loss_closure = get_loss_closure(mll, data_loader=data_loader, **kwargs)
    return ForwardBackwardClosure(forward=loss_closure, parameters=parameters)


@GetLossClosure.register(MarginalLogLikelihood, object, object, DataLoader)
def _get_loss_closure_fallback_external(
    mll: MarginalLogLikelihood,
    _likelihood_type: object,
    _model_type: object,
    data_loader: DataLoader,
    **ignore: Any,
) -> Callable[[], Tensor]:
    r"""Fallback loss closure with externally provided data."""
    batch_generator = chain.from_iterable(iter(data_loader) for _ in repeat(None))

    def closure(**kwargs: Any) -> Tensor:
        batch = next(batch_generator)
        if not isinstance(batch, Sequence):
            raise TypeError(
                "Expected `data_loader` to generate a batch of tensors, "
                f"but found {type(batch)}."
            )

        num_inputs = len(mll.model.train_inputs)
        model_output = mll.model(*batch[:num_inputs])
        log_likelihood = mll(model_output, *batch[num_inputs:], **kwargs)
        return -log_likelihood

    return closure


@GetLossClosure.register(MarginalLogLikelihood, object, object, NoneType)
def _get_loss_closure_fallback_internal(
    mll: MarginalLogLikelihood, _: object, __: object, ___: None, **ignore: Any
) -> Callable[[], Tensor]:
    r"""Fallback loss closure with internally managed data."""

    def closure(**kwargs: Any) -> Tensor:
        model_output = mll.model(*mll.model.train_inputs)
        log_likelihood = mll(model_output, mll.model.train_targets, **kwargs)
        return -log_likelihood

    return closure


@GetLossClosure.register(ExactMarginalLogLikelihood, object, object, NoneType)
def _get_loss_closure_exact_internal(
    mll: ExactMarginalLogLikelihood, _: object, __: object, ___: None, **ignore: Any
) -> Callable[[], Tensor]:
    r"""ExactMarginalLogLikelihood loss closure with internally managed data."""

    def closure(**kwargs: Any) -> Tensor:
        model = mll.model
        # The inputs will get transformed in forward here.
        model_output = model(*model.train_inputs)
        log_likelihood = mll(
            model_output,
            model.train_targets,
            # During model training, the model inputs get transformed in the forward
            # pass. The train_inputs property is not transformed yet, so we need to
            # transform it before passing it to the likelihood for consistency.
            *(model.transform_inputs(X=t_in) for t_in in model.train_inputs),
            **kwargs,
        )
        return -log_likelihood

    return closure


@GetLossClosure.register(SumMarginalLogLikelihood, object, object, NoneType)
def _get_loss_closure_sum_internal(
    mll: SumMarginalLogLikelihood, _: object, __: object, ___: None, **ignore: Any
) -> Callable[[], Tensor]:
    r"""SumMarginalLogLikelihood loss closure with internally managed data."""

    def closure(**kwargs: Any) -> Tensor:
        model = mll.model
        # The inputs will get transformed in forward here.
        model_output = model(*model.train_inputs)
        log_likelihood = mll(
            model_output,
            model.train_targets,
            # During model training, the model inputs get transformed in the forward
            # pass. The train_inputs property is not transformed yet, so we need to
            # transform it before passing it to the likelihood for consistency.
            *(
                (model.transform_inputs(X=t_in) for t_in in sub_t_in)
                for sub_t_in in model.train_inputs
            ),
            **kwargs,
        )
        return -log_likelihood

    return closure



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""Abstract base module for all botorch acquisition functions."""

from __future__ import annotations

from abc import ABC, abstractmethod

import torch
from botorch.acquisition import AcquisitionFunction
from botorch.acquisition.objective import (
    PosteriorTransform,
    ScalarizedPosteriorTransform,
)

from botorch.exceptions.errors import UnsupportedError
from botorch.models.model import Model
from botorch.utils.transforms import (
    average_over_ensemble_models,
    t_batch_mode_transform,
)
from torch import Tensor


class DiscretizedAcquistionFunction(AcquisitionFunction, ABC):
    r"""DiscretizedAcquistionFunction is an abstract base class for acquisition
    functions that are defined on discrete distributions. It wraps a model and
    implements a forward method that computes the acquisition function value at
    a given set of points.
    This class can be subclassed to define acquisiton functions for Riemann-
    distributed posteriors.
    The acquisition function must have the form $$acq(x) = \int p(y|x) ag(x)$$,
    where $$ag$$ is defined differently for each acquisition function.
    The ag_integrate method, which computes the integral of ag between two points, must
    be implemented by subclasses to define the specific acquisition functions.
    """

    def __init__(self, model: Model, posterior_transform: PosteriorTransform) -> None:
        r"""
        Initialize the DiscretizedAcquistionFunction

        Args:
            model: A fitted model that is used to compute the posterior
                distribution over the outcomes of interest.
                The model should be a `PFNModel`.
            posterior_transform: A ScalarizedPosteriorTransform that can only
                indicate minimization or maximization of the objective.
        """
        super().__init__(model=model)
        self.maximize = True
        if posterior_transform is not None:
            unsupported_error_message = (
                "Only scalarized posterior transforms with a"
                "single objective and 0.0 offset are supported."
            )
            if (
                not isinstance(posterior_transform, ScalarizedPosteriorTransform)
                or (posterior_transform.offset != 0.0)
                or len(posterior_transform.weights) != 1
                or posterior_transform.weights[0] not in [-1.0, 1.0]
            ):
                raise UnsupportedError(unsupported_error_message)

            self.maximize = posterior_transform.weights[0] == 1.0

    @t_batch_mode_transform(expected_q=1)
    @average_over_ensemble_models
    def forward(self, X: Tensor) -> Tensor:
        r"""Evaluate the acquisition function on the candidate set X.

        Args:
            X: A `(b) x q x d`-dim Tensor of `(b)` t-batches with `q` `d`-dim
                design points each.

        Returns:
            A `(b)`-dim Tensor of the acquisition function at the given
            design points `X`.
        """
        discrete_posterior = self.model.posterior(X)
        if not self.maximize:
            discrete_posterior.borders = -torch.flip(discrete_posterior.borders, [0])
            discrete_posterior.probabilities = torch.flip(
                discrete_posterior.probabilities, [-1]
            )

        result = discrete_posterior.integrate(self.ag_integrate)
        # remove q dimension
        return result.squeeze(-1)

    @abstractmethod
    def ag_integrate(self, lower_bound: Tensor, upper_bound: Tensor) -> Tensor:
        r"""
        This function calculates the integral that computes the acquisition function
        without the posterior factor from lower_bound to upper_bound.
        That is, our acquisition function is assumed to have the form
        \int ag(x) * p(x) dx,
        and this function calculates \int_{lower_bound}^{upper_bound} ag(x) dx.
        The `integrate` method of the posterior (`BoundedRiemannPosterior`)
        then computes the final acquisition value.

        Args:
            lower_bound: lower bound of integral
            upper_bound: upper bound of integral

        Returns:
            A `(b)`-dim Tensor of acquisition function derivatives at the given
            design points `X`.
        """
        pass  # pragma: no cover


class DiscretizedExpectedImprovement(DiscretizedAcquistionFunction):
    r"""DiscretizedExpectedImprovement is an acquisition function that
    computes the expected improvement over the current best observed value
    for a Riemann distribution.
    """

    def __init__(
        self,
        model: Model,
        best_f: Tensor,
        posterior_transform: PosteriorTransform | None = None,
    ) -> None:
        r"""
        Initialize the DiscretizedExpectedImprovement

        Args:
            model: A fitted model that is used to compute the posterior
                distribution over the outcomes of interest.
                The model should be a `PFNModel`.
            best_f: A tensor representing the current best observed value.
        """
        super().__init__(model=model, posterior_transform=posterior_transform)
        self.register_buffer("best_f", torch.as_tensor(best_f))

    def ag_integrate(self, lower_bound: Tensor, upper_bound: Tensor) -> Tensor:
        r"""
        As Expected improvement has ag(y) = (y - best_f).clamp(min=0), and
        is defined as \int ag(y) * p(y) dy, we can calculate the integral
        of ag(y) like so:
        We just calculate ag(y) at beginning and end, and since the function has
        a gradient of 1 or 0, we can just take the average of the two.

        Args:
            lower_bound: lower bound of integral
            upper_bound: upper bound of integral

        Returns:
            A `(b)`-dim Tensor of acquisition function derivatives at the given
            design points `X`.
        """
        best_f = self.best_f.to(lower_bound)

        # Case 1: best_f >= upper_bound, entire interval gives 0 improvement
        case1_mask = best_f >= upper_bound

        # Case 2: best_f <= lower_bound, entire interval gives improvement
        case2_mask = best_f <= lower_bound

        # Case 3: lower_bound < best_f < upper_bound, partial improvement
        case3_mask = ~(case1_mask | case2_mask)

        # Initialize result tensor
        result = torch.zeros_like(lower_bound)

        # Case 1: result is already 0

        # Case 2: integral = (
        #    ((upper_bound + lower_bound)/2 - best_f)
        #     * (upper_bound - lower_bound)
        # )
        if case2_mask.any():
            bucket_width = upper_bound - lower_bound
            bucket_center = (upper_bound + lower_bound) / 2
            result = torch.where(
                case2_mask, (bucket_center - best_f) * bucket_width, result
            )

        # Case 3: integral = (upper_bound - best_f)²/2
        if case3_mask.any():
            result = torch.where(case3_mask, (upper_bound - best_f).pow(2) / 2, result)

        return result.clamp_min(0)


class DiscretizedProbabilityOfImprovement(DiscretizedAcquistionFunction):
    r"""DiscretizedProbabilityOfImprovement is an acquisition function that
    computes the probability of improvement over the current best observed value
    for a Riemann distribution.
    """

    def __init__(
        self,
        model: Model,
        best_f: Tensor,
        posterior_transform: PosteriorTransform | None = None,
    ) -> None:
        r"""
        Initialize the DiscretizedProbabilityOfImprovement

        Args:
            model: A fitted model that is used to compute the posterior
                distribution over the outcomes of interest.
                The model should be a `PFNModel`.
            best_f: A tensor representing the current best observed value.
        """

        super().__init__(model, posterior_transform)
        self.register_buffer("best_f", torch.as_tensor(best_f))

    def ag_integrate(self, lower_bound: Tensor, upper_bound: Tensor) -> Tensor:
        r"""
        PI is defined as \int ag(y) * p(y) dy, where ag(y) = I(y - best_f)
        and I being the indicator function.

        So all we need to do is calculate the portion between the bounds
        that is larger than best_f.
        We do this by comparing how much higher the upper bound is than best_f,
        compared to the size of the bucket.
        Then we clamp at one if best_f is below lower_bound and at zero if
        best_f is above upper_bound.

        Args:
            lower_bound: lower bound of integral
            upper_bound: upper bound of integral

        Returns:
            A `(b)`-dim Tensor of acquisition function derivatives at the given
            design points `X`.
        """
        best_f = self.best_f.to(lower_bound)
        # two separate clamps needed below, as one is a tensor and one a scalar
        return (
            (upper_bound - best_f).clamp(min=0.0).clamp(max=upper_bound - lower_bound)
        )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
Discretization (rounding) functions for acquisition optimization.

References

.. [Daulton2022bopr]
    S. Daulton, X. Wan, D. Eriksson, M. Balandat, M. A. Osborne, E. Bakshy.
    Bayesian Optimization over Discrete and Mixed Spaces via Probabilistic
    Reparameterization. Advances in Neural Information Processing Systems
    35, 2022.
"""

from __future__ import annotations

import torch
from torch import Tensor
from torch.autograd import Function
from torch.nn.functional import one_hot


def approximate_round(X: Tensor, tau: float = 1e-3) -> Tensor:
    r"""Diffentiable approximate rounding function.

    This method is a piecewise approximation of a rounding function where
    each piece is a hyperbolic tangent function.

    Args:
        X: The tensor to round to the nearest integer (element-wise).
        tau: A temperature hyperparameter.

    Returns:
        The approximately rounded input tensor.
    """
    offset = X.floor()
    scaled_remainder = (X - offset - 0.5) / tau
    rounding_component = (torch.tanh(scaled_remainder) + 1) / 2
    return offset + rounding_component


class IdentitySTEFunction(Function):
    """Base class for functions using straight through gradient estimators.

    This class approximates the gradient with the identity function.
    """

    @staticmethod
    def backward(ctx, grad_output: Tensor) -> Tensor:
        r"""Use a straight-through estimator the gradient.

        This uses the identity function.

        Args:
            grad_output: A tensor of gradients.

        Returns:
            The provided tensor.
        """
        return grad_output


class RoundSTE(IdentitySTEFunction):
    r"""Round the input tensor and use a straight-through gradient estimator.

    [Daulton2022bopr]_ proposes using this in acquisition optimization.
    """

    @staticmethod
    def forward(ctx, X: Tensor) -> Tensor:
        r"""Round the input tensor element-wise.

        Args:
            X: The tensor to be rounded.

        Returns:
            A tensor where each element is rounded to the nearest integer.
        """
        return X.round()


class OneHotArgmaxSTE(IdentitySTEFunction):
    r"""Discretize a continuous relaxation of a one-hot encoded categorical.

    This returns a one-hot encoded categorical and use a straight-through
    gradient estimator via an identity function.

    [Daulton2022bopr]_ proposes using this in acquisition optimization.
    """

    @staticmethod
    def forward(ctx, X: Tensor) -> Tensor:
        r"""Discretize the input tensor.

        This applies a argmax along the last dimensions of the input tensor
        and one-hot encodes the result.

        Args:
            X: The tensor to be rounded.

        Returns:
            A tensor where each element is rounded to the nearest integer.
        """
        return one_hot(X.argmax(dim=-1), num_classes=X.shape[-1]).to(X)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import annotations

import torch
from gpytorch.constraints import Positive
from gpytorch.kernels import Kernel
from torch import Tensor


class InfiniteWidthBNNKernel(Kernel):
    r"""Infinite-width BNN kernel.

    Defines the GP kernel which is equivalent to performing exact Bayesian
    inference on a fully-connected deep neural network with ReLU activations
    and i.i.d. priors in the infinite-width limit.
    See [Cho2009kernel]_ and [Lee2018deep]_ for details.

    .. [Cho2009kernel]
        Y. Cho, and L. Saul. Kernel methods for deep learning.
        Advances in Neural Information Processing Systems 22. 2009.
    .. [Lee2018deep]
        J. Lee, Y. Bahri, R. Novak, S. Schoenholz, J. Pennington, and J. Dickstein.
        Deep Neural Networks as Gaussian Processes.
        International Conference on Learning Representations. 2018.
    """

    has_lengthscale = False

    def __init__(
        self,
        depth: int = 3,
        batch_shape: torch.Size | None = None,
        active_dims: tuple[int, ...] | None = None,
        acos_eps: float = 1e-7,
        device: torch.device | None = None,
    ) -> None:
        r"""
        Args:
            depth: Depth of neural network.
            batch_shape: This will set a separate weight/bias var for each batch.
                It should be :math:`B_1 \times \ldots \times B_k` if :math:`\mathbf` is
                a :math:`B_1 \times \ldots \times B_k \times N \times D` tensor.
            param active_dims: Compute the covariance of only a few input dimensions.
                The ints corresponds to the indices of the dimensions.
            param acos_eps: A small positive value to restrict acos inputs to
                :math`[-1 + \epsilon, 1 - \epsilon]`
            param device: Device for parameters.
        """
        super().__init__(batch_shape=batch_shape, active_dims=active_dims)
        self.depth = depth
        self.acos_eps = acos_eps

        self.register_parameter(
            "raw_weight_var",
            torch.nn.Parameter(torch.zeros(*self.batch_shape, 1, 1, device=device)),
        )
        self.register_constraint("raw_weight_var", Positive())

        self.register_parameter(
            "raw_bias_var",
            torch.nn.Parameter(torch.zeros(*self.batch_shape, 1, 1, device=device)),
        )
        self.register_constraint("raw_bias_var", Positive())

    @property
    def weight_var(self) -> Tensor:
        return self.raw_weight_var_constraint.transform(self.raw_weight_var)

    @weight_var.setter
    def weight_var(self, value) -> None:
        if not torch.is_tensor(value):
            value = torch.as_tensor(value).to(self.raw_weight_var)
        self.initialize(
            raw_weight_var=self.raw_weight_var_constraint.inverse_transform(value)
        )

    @property
    def bias_var(self) -> Tensor:
        return self.raw_bias_var_constraint.transform(self.raw_bias_var)

    @bias_var.setter
    def bias_var(self, value) -> None:
        if not torch.is_tensor(value):
            value = torch.as_tensor(value).to(self.raw_bias_var)
        self.initialize(
            raw_bias_var=self.raw_bias_var_constraint.inverse_transform(value)
        )

    def _initialize_var(self, x: Tensor) -> Tensor:
        """Computes the initial variance of x for layer 0"""
        return (
            self.weight_var * torch.sum(x * x, dim=-1, keepdim=True) / x.shape[-1]
            + self.bias_var
        )

    def _update_var(self, K: Tensor, x: Tensor) -> Tensor:
        """Computes the updated variance of x for next layer"""
        return self.weight_var * K / 2 + self.bias_var

    def k(self, x1: Tensor, x2: Tensor) -> Tensor:
        r"""
        For single-layer infinite-width neural networks with i.i.d. priors,
        the covariance between outputs can be computed by
        :math:`K^0(x, x')=\sigma_b^2+\sigma_w^2\frac{x \cdot x'}{d_\text{input}}`.

        For deeper networks, we can recursively define the covariance as
        :math:`K^l(x, x')=\sigma_b^2+\sigma_w^2
        F_\phi(K^{l-1}(x, x'), K^{l-1}(x, x), K^{l-1}(x', x'))`
        where :math:`F_\phi` is a deterministic function based on the
        activation function :math:`\phi`.

        For ReLU activations, this yields the arc-cosine kernel, which can be computed
        analytically.

        Args:
            x1: `batch_shape x n1 x d`-dim Tensor
            x2: `batch_shape x n2 x d`-dim Tensor
        """
        K_12 = (
            self.weight_var * (x1.matmul(x2.transpose(-2, -1)) / x1.shape[-1])
            + self.bias_var
        )

        for layer in range(self.depth):
            if layer == 0:
                K_11 = self._initialize_var(x1)
                K_22 = self._initialize_var(x2)
            else:
                K_11 = self._update_var(K_11, x1)
                K_22 = self._update_var(K_22, x2)

            sqrt_term = torch.sqrt(K_11.matmul(K_22.transpose(-2, -1)))

            fraction = K_12 / sqrt_term
            fraction = torch.clamp(
                fraction, min=-1 + self.acos_eps, max=1 - self.acos_eps
            )

            theta = torch.acos(fraction)
            theta_term = torch.sin(theta) + (torch.pi - theta) * fraction

            K_12 = (
                self.weight_var / (2 * torch.pi) * sqrt_term * theta_term
                + self.bias_var
            )

        return K_12

    def forward(
        self,
        x1: Tensor,
        x2: Tensor,
        diag: bool | None = False,
        last_dim_is_batch: bool | None = False,
        **params,
    ) -> Tensor:
        """
        Args:
            x1: `batch_shape x n1 x d`-dim Tensor
            x2: `batch_shape x n2 x d`-dim Tensor
            diag: If True, only returns the diagonal of the kernel matrix.
            last_dim_is_batch: Not supported by this kernel.
        """
        if last_dim_is_batch:
            raise RuntimeError("last_dim_is_batch not supported by this kernel.")

        if diag:
            K = self._initialize_var(x1)
            for _ in range(self.depth):
                K = self._update_var(K, x1)
            return K.squeeze(-1)
        else:
            return self.k(x1, x2)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import annotations

from collections.abc import Callable

import torch
from botorch.posteriors.posterior import Posterior
from torch import Tensor


class TransformedPosterior(Posterior):
    r"""A generic transformation of a posterior (implicitly represented)."""

    def __init__(
        self,
        posterior: Posterior,
        sample_transform: Callable[[Tensor], Tensor],
        mean_transform: Callable[[Tensor, Tensor], Tensor] | None = None,
        variance_transform: Callable[[Tensor, Tensor], Tensor] | None = None,
    ) -> None:
        r"""An implicitly represented transformed posterior.

        Args:
            posterior: The posterior object to be transformed.
            sample_transform: A callable applying a sample-level transform to a
                `sample_shape x batch_shape x q x m`-dim tensor of samples from
                the original posterior, returning a tensor of samples of the
                same shape.
            mean_transform: A callable transforming a 2-tuple of mean and
                variance (both of shape `batch_shape x m x o`) of the original
                posterior to the mean of the transformed posterior.
            variance_transform: A callable transforming a 2-tuple of mean and
                variance (both of shape `batch_shape x m x o`) of the original
                posterior to a variance of the transformed posterior.
        """
        self._posterior = posterior
        self._sample_transform = sample_transform
        self._mean_transform = mean_transform
        self._variance_transform = variance_transform

    @property
    def base_sample_shape(self) -> torch.Size:
        r"""The shape of a base sample used for constructing posterior samples."""
        return self._posterior.base_sample_shape

    @property
    def batch_range(self) -> tuple[int, int]:
        r"""The t-batch range.

        This is used in samplers to identify the t-batch component of the
        `base_sample_shape`. The base samples are expanded over the t-batches to
        provide consistency in the acquisition values, i.e., to ensure that a
        candidate produces same value regardless of its position on the t-batch.
        """
        return self._posterior.batch_range

    @property
    def device(self) -> torch.device:
        r"""The torch device of the posterior."""
        return self._posterior.device

    @property
    def dtype(self) -> torch.dtype:
        r"""The torch dtype of the posterior."""
        return self._posterior.dtype

    def _extended_shape(
        self,
        sample_shape: torch.Size = torch.Size(),  # noqa: B008
    ) -> torch.Size:
        r"""Returns the shape of the samples produced by the posterior with
        the given `sample_shape`.

        NOTE: This assumes that the `sample_transform` does not change the
        shape of the samples.
        """
        return self._posterior._extended_shape(sample_shape=sample_shape)

    @property
    def mean(self) -> Tensor:
        r"""The mean of the posterior as a `batch_shape x n x m`-dim Tensor."""
        if self._mean_transform is None:
            raise NotImplementedError("No mean transform provided.")
        try:
            variance = self._posterior.variance
        except (NotImplementedError, AttributeError):
            variance = None
        return self._mean_transform(self._posterior.mean, variance)

    @property
    def variance(self) -> Tensor:
        r"""The variance of the posterior as a `batch_shape x n x m`-dim Tensor."""
        if self._variance_transform is None:
            raise NotImplementedError("No variance transform provided.")
        return self._variance_transform(self._posterior.mean, self._posterior.variance)

    def rsample_from_base_samples(
        self,
        sample_shape: torch.Size,
        base_samples: Tensor,
    ) -> Tensor:
        r"""Sample from the posterior (with gradients) using base samples.

        This is intended to be used with a sampler that produces the corresponding base
        samples, and enables acquisition optimization via Sample Average Approximation.

        Args:
            sample_shape: A `torch.Size` object specifying the sample shape. To
                draw `n` samples, set to `torch.Size([n])`. To draw `b` batches
                of `n` samples each, set to `torch.Size([b, n])`.
            base_samples: The base samples, obtained from the appropriate sampler.
                This is a tensor of shape `sample_shape x base_sample_shape`.

        Returns:
            Samples from the posterior, a tensor of shape
            `self._extended_shape(sample_shape=sample_shape)`.
        """
        samples = self._posterior.rsample_from_base_samples(
            sample_shape=sample_shape, base_samples=base_samples
        )
        return self._sample_transform(samples)

    def rsample(
        self,
        sample_shape: torch.Size | None = None,
    ) -> Tensor:
        r"""Sample from the posterior (with gradients).

        Args:
            sample_shape: A `torch.Size` object specifying the sample shape. To
                draw `n` samples, set to `torch.Size([n])`. To draw `b` batches
                of `n` samples each, set to `torch.Size([b, n])`.

        Returns:
            Samples from the posterior, a tensor of shape
            `self._extended_shape(sample_shape=sample_shape)`.
        """
        samples = self._posterior.rsample(sample_shape=sample_shape)
        return self._sample_transform(samples)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
A registry of helpers for generating inputs to acquisition function
constructors programmatically from a consistent input format.

Contributor: hvarfner (bayesian_active_learning, scorebo)
"""

from __future__ import annotations

from typing import Any, Hashable, List, Optional, Tuple

import torch

from botorch.acquisition.input_constructors import (
    acqf_input_constructor,
    get_best_f_analytic,
)
from botorch.acquisition.objective import (
    PosteriorTransform,
    ScalarizedPosteriorTransform,
)
from botorch.acquisition.utils import get_optimal_samples
from botorch.models.model import Model

from botorch.utils.datasets import SupervisedDataset
from botorch_community.acquisition.bayesian_active_learning import (
    qBayesianQueryByComittee,
    qBayesianVarianceReduction,
    qStatisticalDistanceActiveLearning,
)

from botorch_community.acquisition.discretized import (
    DiscretizedExpectedImprovement,
    DiscretizedProbabilityOfImprovement,
)
from botorch_community.acquisition.scorebo import qSelfCorrectingBayesianOptimization
from torch import Tensor


@acqf_input_constructor(
    DiscretizedExpectedImprovement, DiscretizedProbabilityOfImprovement
)
def construct_inputs_best_f(
    model: Model,
    training_data: SupervisedDataset | dict[Hashable, SupervisedDataset],
    posterior_transform: PosteriorTransform | None = None,
    best_f: float | Tensor | None = None,
) -> dict[str, Any]:
    r"""Construct kwargs for the acquisition functions requiring `best_f`.

    Args:
        model: The model to be used in the acquisition function.
        training_data: Dataset(s) used to train the model.
            Used to determine default value for `best_f`.
        best_f: Threshold above (or below) which improvement is defined.
        posterior_transform: The posterior transform to be used in the
            acquisition function.

    Returns:
        A dict mapping kwarg names of the constructor to values.
    """
    if best_f is None:
        best_f = get_best_f_analytic(
            training_data=training_data,
            posterior_transform=posterior_transform,
        )

    return {
        "model": model,
        "posterior_transform": posterior_transform,
        "best_f": best_f,
    }


@acqf_input_constructor(
    qBayesianQueryByComittee,
    qBayesianVarianceReduction,
)
def construct_inputs_BAL(
    model: Model,
    X_pending: Optional[Tensor] = None,
):
    inputs = {
        "model": model,
        "X_pending": X_pending,
    }
    return inputs


@acqf_input_constructor(qStatisticalDistanceActiveLearning)
def construct_inputs_SAL(
    model: Model,
    distance_metric: str = "hellinger",
    X_pending: Optional[Tensor] = None,
):
    inputs = {
        "model": model,
        "distance_metric": distance_metric,
        "X_pending": X_pending,
    }
    return inputs


@acqf_input_constructor(qSelfCorrectingBayesianOptimization)
def construct_inputs_SCoreBO(
    model: Model,
    bounds: List[Tuple[float, float]],
    num_optima: int = 8,
    posterior_transform: Optional[ScalarizedPosteriorTransform] = None,
    distance_metric: str = "hellinger",
    X_pending: Optional[Tensor] = None,
):
    dtype = model.train_targets.dtype
    # the number of optima are per model
    optimal_inputs, optimal_outputs = get_optimal_samples(
        model=model,
        bounds=torch.as_tensor(bounds, dtype=dtype).T,
        num_optima=num_optima,
        posterior_transform=posterior_transform,
        return_transformed=True,
    )
    inputs = {
        "model": model,
        "optimal_inputs": optimal_inputs,
        "optimal_outputs": optimal_outputs,
        "distance_metric": distance_metric,
        "posterior_transform": posterior_transform,
        "X_pending": X_pending,
    }
    return inputs



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""Core methods for building closures in torch and interfacing with numpy."""

from __future__ import annotations

from collections.abc import Callable, Sequence

from typing import Any

import numpy as np
import numpy.typing as npt
import torch

from botorch.optim.utils import _handle_numerical_errors
from botorch.optim.utils.numpy_utils import as_ndarray
from botorch.utils.context_managers import zero_grad_ctx
from numpy import float64 as np_float64, zeros as np_zeros
from torch import Tensor

FILL_VALUE = 0.0


class ForwardBackwardClosure:
    r"""Wrapper for fused forward and backward closures."""

    def __init__(
        self, forward: Callable[[], Tensor], parameters: dict[str, Tensor]
    ) -> None:
        r"""Initializes a ForwardBackwardClosure instance.

        Args:
            forward: Callable that returns a tensor.
            parameters: A dictionary of tensors whose `grad` fields are to be returned.
        """
        self.forward = forward
        self.parameters = parameters

    def __call__(self, **kwargs: Any) -> tuple[Tensor, tuple[Tensor | None, ...]]:
        with zero_grad_ctx(parameters=self.parameters):
            value = self.forward(**kwargs).sum()
            value.backward()

            grads = tuple(param.grad for param in self.parameters.values())
            return value, grads


class NdarrayOptimizationClosure:
    r"""Adds stateful behavior and a numpy.ndarray-typed API to a closure with an
    expected return type Tuple[Tensor, Union[Tensor, Sequence[Optional[Tensor]]]].

    NaN values will be replaced with 0.0 in the returned ndarray."""

    def __init__(
        self,
        closure: Callable[[], tuple[Tensor, Sequence[Tensor | None]]],
        parameters: dict[str, Tensor],
    ) -> None:
        r"""Initializes a NdarrayOptimizationClosure instance.

        Args:
            closure: A ForwardBackwardClosure instance.
            parameters: A dictionary of tensors representing the closure's state.
                Expected to correspond with the first `len(parameters)` optional
                gradient tensors returned by `closure`.
        """

        self.closure = closure
        self.parameters = parameters
        self._gradient_ndarray: npt.NDArray | None = None

    def __call__(
        self, state: npt.NDArray | None = None, **kwargs: Any
    ) -> tuple[npt.NDArray, npt.NDArray]:
        if state is not None:
            self.state = state

        try:
            value_tensor, grad_tensors = self.closure(**kwargs)
            value = as_ndarray(values=value_tensor, dtype=np_float64)
            grads = self._get_gradient_ndarray()
            index = 0
            for param, grad in zip(self.parameters.values(), grad_tensors):
                size = param.numel()
                if grad is not None:
                    grads[index : index + size] = as_ndarray(
                        values=grad.view(-1), dtype=np_float64
                    )
                index += size
        except RuntimeError as e:
            value, grads = _handle_numerical_errors(e, x=self.state, dtype=np_float64)

        return value, grads

    @property
    def state(self) -> npt.NDArray:
        if len(self.parameters) == 0:
            raise RuntimeError("No parameters to get state from.")

        size = sum(tnsr.numel() for tnsr in self.parameters.values())
        tnsr = next(iter(self.parameters.values()))
        dtype = np_float64

        out = np.empty([size], dtype=dtype)

        index = 0
        for tnsr in self.parameters.values():
            size = tnsr.numel()
            out[index : index + size] = as_ndarray(tnsr.view(-1))
            index += size

        return out

    @state.setter
    def state(self, state: npt.NDArray) -> None:
        with torch.no_grad():
            index = 0
            for tnsr in self.parameters.values():
                size = tnsr.numel()
                vals = state[index : index + size] if tnsr.ndim else state[index]
                tnsr.copy_(
                    torch.as_tensor(vals, device=tnsr.device, dtype=tnsr.dtype).view(
                        tnsr.shape
                    )
                )
                index += size

    def _get_gradient_ndarray(self) -> npt.NDArray:
        if self._gradient_ndarray is not None:
            self._gradient_ndarray.fill(FILL_VALUE)
            return self._gradient_ndarray

        size = sum(param.numel() for param in self.parameters.values())
        self._gradient_ndarray = np_zeros(size, dtype=np_float64)
        return self._gradient_ndarray



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
Acquisition function for joint entropy search (JES).

.. [Hvarfner2022joint]
    C. Hvarfner, F. Hutter, L. Nardi,
    Joint Entropy Search for Maximally-informed Bayesian Optimization.
    In Proceedings of the Annual Conference on Neural Information
    Processing Systems (NeurIPS), 2022.

.. [Tu2022joint]
    B. Tu, A. Gandy, N. Kantas, B. Shafei,
    Joint Entropy Search for Multi-objective Bayesian Optimization.
    In Proceedings of the Annual Conference on Neural Information
    Processing Systems (NeurIPS), 2022.
"""

from __future__ import annotations

import warnings
from math import log, pi

import torch
from botorch import settings
from botorch.acquisition.acquisition import AcquisitionFunction, MCSamplerMixin
from botorch.acquisition.objective import PosteriorTransform
from botorch.models.fully_bayesian import FullyBayesianSingleTaskGP
from botorch.models.model import Model
from botorch.models.utils import check_no_nans, fantasize as fantasize_flag
from botorch.models.utils.gpytorch_modules import MIN_INFERRED_NOISE_LEVEL
from botorch.sampling.normal import SobolQMCNormalSampler
from botorch.utils.transforms import (
    average_over_ensemble_models,
    concatenate_pending_points,
    t_batch_mode_transform,
)
from torch import Tensor

from torch.distributions import Normal

MCMC_DIM = -3  # Only relevant if you do Fully Bayesian GPs.
ESTIMATION_TYPES = ["MC", "LB"]
MC_ADD_TERM = 0.5 * (1 + log(2 * pi))

# The CDF query cannot be strictly zero in the division
# and this clamping helps assure that it is always positive.
CLAMP_LB = torch.finfo(torch.float32).eps
FULLY_BAYESIAN_ERROR_MSG = (
    "JES is not yet available with Fully Bayesian GPs. Track the issue, "
    "which regards conditioning on a number of optima on a collection "
    "of models, in detail at https://github.com/pytorch/botorch/issues/1680"
)


class qJointEntropySearch(AcquisitionFunction, MCSamplerMixin):
    r"""The acquisition function for the Joint Entropy Search, where the batches
    `q > 1` are supported through the lower bound formulation.

    This acquisition function computes the mutual information between the observation
    at a candidate point `X` and the optimal input-output pair.

    See [Tu2022joint]_ for a discussion on the estimation procedure.
    """

    def __init__(
        self,
        model: Model,
        optimal_inputs: Tensor,
        optimal_outputs: Tensor,
        condition_noiseless: bool = True,
        posterior_transform: PosteriorTransform | None = None,
        X_pending: Tensor | None = None,
        estimation_type: str = "LB",
        num_samples: int = 64,
    ) -> None:
        r"""Joint entropy search acquisition function.

        Args:
            model: A fitted single-outcome model.
            optimal_inputs: A `num_samples x d`-dim tensor containing the sampled
                optimal inputs of dimension `d`. We assume for simplicity that each
                sample only contains one optimal set of inputs.
            optimal_outputs: A `num_samples x 1`-dim Tensor containing the optimal
                set of objectives of dimension `1`.
            condition_noiseless: Whether to condition on noiseless optimal observations
                `f*` [Hvarfner2022joint]_ or noisy optimal observations `y*`
                [Tu2022joint]_. These are sampled identically, so this only controls
                the fashion in which the GP is reshaped as a result of conditioning
                on the optimum.
            posterior_transform: PosteriorTransform to negate or scalarize the output.
            estimation_type: estimation_type: A string to determine which entropy
                estimate is computed: Lower bound" ("LB") or "Monte Carlo" ("MC").
                Lower Bound is recommended due to the relatively high variance
                of the MC estimator.
            X_pending: A `m x d`-dim Tensor of `m` design points that have been
                submitted for function evaluation, but have not yet been evaluated.
            num_samples: The number of Monte Carlo samples used for the Monte Carlo
                estimate.
        """
        super().__init__(model=model)
        sampler = SobolQMCNormalSampler(sample_shape=torch.Size([num_samples]))
        MCSamplerMixin.__init__(self, sampler=sampler)
        # To enable fully bayesian GP conditioning, we need to unsqueeze
        # to get num_optima x num_gps unique GPs

        # inputs come as num_optima_per_model x (num_models) x d
        # but we want it four-dimensional in the Fully bayesian case,
        # and three-dimensional otherwise.
        self.optimal_inputs = optimal_inputs.unsqueeze(-2)
        self.optimal_outputs = optimal_outputs.unsqueeze(-2)
        self.optimal_output_values = (
            posterior_transform.evaluate(self.optimal_outputs).unsqueeze(-1)
            if posterior_transform
            else self.optimal_outputs
        )
        self.posterior_transform = posterior_transform

        self.num_samples = optimal_inputs.shape[0]
        self.condition_noiseless = condition_noiseless
        self.initial_model = model

        # Here, the optimal inputs have shapes num_optima x [num_models if FB] x 1 x D
        # and the optimal outputs have shapes num_optima x [num_models if FB] x 1 x 1
        # The third dimension equaling 1 is required to get one optimum per model,
        # which raises a BotorchTensorDimensionWarning.
        if isinstance(model, FullyBayesianSingleTaskGP):
            raise NotImplementedError(FULLY_BAYESIAN_ERROR_MSG)
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore")
            with fantasize_flag():
                with settings.propagate_grads(False):
                    # We must do a forward pass one before conditioning.
                    self.initial_model.posterior(
                        self.optimal_inputs[:1], observation_noise=False
                    )

                # This equates to the JES version proposed by Hvarfner et. al.
                if self.condition_noiseless:
                    opt_noise = torch.full_like(
                        self.optimal_outputs, MIN_INFERRED_NOISE_LEVEL
                    )
                    # conditional (batch) model of shape (num_models)
                    # x num_optima_per_model
                    self.conditional_model = (
                        self.initial_model.condition_on_observations(
                            X=self.initial_model.transform_inputs(self.optimal_inputs),
                            Y=self.optimal_outputs,
                            noise=opt_noise,
                        )
                    )
                else:
                    self.conditional_model = (
                        self.initial_model.condition_on_observations(
                            X=self.initial_model.transform_inputs(self.optimal_inputs),
                            Y=self.optimal_outputs,
                        )
                    )

        self.estimation_type = estimation_type
        self.set_X_pending(X_pending)

    @concatenate_pending_points
    @t_batch_mode_transform()
    @average_over_ensemble_models
    def forward(self, X: Tensor) -> Tensor:
        r"""Evaluates qJointEntropySearch at the design points `X`.

        Args:
            X: A `batch_shape x q x d`-dim Tensor of `batch_shape` t-batches with `q`
            `d`-dim design points each.

        Returns:
            A `batch_shape`-dim Tensor of acquisition values at the given design
            points `X`.
        """
        if self.estimation_type == "LB":
            res = self._compute_lower_bound_information_gain(X)
        elif self.estimation_type == "MC":
            res = self._compute_monte_carlo_information_gain(X)
        else:
            raise ValueError(
                f"Estimation type {self.estimation_type} is not valid. "
                f"Please specify any of {ESTIMATION_TYPES}"
            )
        return res

    def _compute_lower_bound_information_gain(
        self, X: Tensor, return_parts: bool = False
    ) -> Tensor:
        r"""Evaluates the lower bound information gain at the design points `X`.

        Args:
            X: A `batch_shape x q x d`-dim Tensor of `batch_shape` t-batches with `q`
            `d`-dim design points each.

        Returns:
            A `batch_shape`-dim Tensor of acquisition values at the given design
            points `X`.
        """
        initial_posterior = self.initial_model.posterior(
            X, observation_noise=True, posterior_transform=self.posterior_transform
        )
        # need to check if there is a two-dimensional batch shape -
        # the sampled optima appear in the dimension right after
        batch_shape = X.shape[:-2]
        sample_dim = len(batch_shape)
        # We DISREGARD the additional constant term.
        initial_entropy = 0.5 * torch.logdet(
            initial_posterior.mvn.lazy_covariance_matrix
        )

        # initial_entropy of shape batch_size or batch_size x num_models if FBGP
        # first need to unsqueeze the sample dim (after batch dim) and then the two last
        initial_entropy = (
            initial_entropy.unsqueeze(sample_dim).unsqueeze(-1).unsqueeze(-1)
        )

        # Compute the mixture mean and variance
        posterior_m = self.conditional_model.posterior(
            X.unsqueeze(MCMC_DIM),
            observation_noise=True,
            posterior_transform=self.posterior_transform,
        )
        noiseless_var = self.conditional_model.posterior(
            X.unsqueeze(MCMC_DIM),
            observation_noise=False,
            posterior_transform=self.posterior_transform,
        ).variance

        mean_m = posterior_m.mean
        variance_m = posterior_m.variance

        check_no_nans(variance_m)
        # get stdv of noiseless variance
        stdv = noiseless_var.sqrt()
        # batch_shape x 1
        normal = Normal(
            torch.zeros(1, device=X.device, dtype=X.dtype),
            torch.ones(1, device=X.device, dtype=X.dtype),
        )
        normalized_mvs = (self.optimal_output_values - mean_m) / stdv
        cdf_mvs = normal.cdf(normalized_mvs).clamp_min(CLAMP_LB)
        pdf_mvs = torch.exp(normal.log_prob(normalized_mvs))

        ratio = pdf_mvs / cdf_mvs
        var_truncated = noiseless_var * (
            1 - (normalized_mvs + ratio) * ratio
        ).clamp_min(CLAMP_LB)

        var_truncated = var_truncated + (variance_m - noiseless_var)
        conditional_entropy = 0.5 * torch.log(var_truncated)

        # Shape batch_size x num_optima x [num_models if FB] x q x num_outputs
        # squeeze the num_outputs dim (since it's 1)
        entropy_reduction = (
            initial_entropy - conditional_entropy.sum(dim=-2, keepdim=True)
        ).squeeze(-1)
        # average over the number of optima and squeeze the q-batch

        entropy_reduction = entropy_reduction.mean(dim=sample_dim).squeeze(-1)
        return entropy_reduction

    def _compute_monte_carlo_variables(self, posterior):
        """Retrieves monte carlo samples and their log probabilities from the posterior.

        Args:
            posterior: The posterior distribution.

        Returns:
            A two-element tuple containing:
            - samples: a num_optima x batch_shape x num_mc_samples x q x 1
                tensor of samples drawn from the posterior.
            - samples_log_prob: a num_optima x batch_shape x num_mc_samples x q x 1
                tensor of associated probabilities.
        """
        samples = self.get_posterior_samples(posterior)
        samples_log_prob = (
            posterior.mvn.log_prob(samples.squeeze(-1)).unsqueeze(-1).unsqueeze(-1)
        )
        return samples, samples_log_prob

    def _compute_monte_carlo_information_gain(
        self, X: Tensor, return_parts: bool = False
    ) -> Tensor:
        r"""Evaluates the lower bound information gain at the design points `X`.

        Args:
            X: A `batch_shape x q x d`-dim Tensor of `batch_shape` t-batches with `q`
            `d`-dim design points each.

        Returns:
            A `batch_shape`-dim Tensor of acquisition values at the given design
            points `X`.
        """
        initial_posterior = self.initial_model.posterior(
            X, observation_noise=True, posterior_transform=self.posterior_transform
        )

        batch_shape = X.shape[:-2]
        sample_dim = len(batch_shape)
        # We DISREGARD the additional constant term.
        initial_entropy = MC_ADD_TERM + 0.5 * torch.logdet(
            initial_posterior.mvn.lazy_covariance_matrix
        )

        # initial_entropy of shape batch_size or batch_size x num_models if FBGP
        # first need to unsqueeze the sample dim (after batch dim), then the two last
        initial_entropy = (
            initial_entropy.unsqueeze(sample_dim).unsqueeze(-1).unsqueeze(-1)
        )

        # Compute the mixture mean and variance
        posterior_m = self.conditional_model.posterior(
            X.unsqueeze(MCMC_DIM),
            observation_noise=True,
            posterior_transform=self.posterior_transform,
        )
        noiseless_var = self.conditional_model.posterior(
            X.unsqueeze(MCMC_DIM),
            observation_noise=False,
            posterior_transform=self.posterior_transform,
        ).variance

        mean_m = posterior_m.mean
        variance_m = posterior_m.variance.clamp_min(CLAMP_LB)
        conditional_samples, conditional_logprobs = self._compute_monte_carlo_variables(
            posterior_m
        )

        normalized_samples = (conditional_samples - mean_m) / variance_m.sqrt()
        # Correlation between noisy observations and noiseless values f
        rho = (noiseless_var / variance_m).sqrt()

        normal = Normal(
            torch.zeros(1, device=X.device, dtype=X.dtype),
            torch.ones(1, device=X.device, dtype=X.dtype),
        )
        # prepare max value quantities and re-scale as required
        normalized_mvs = (self.optimal_outputs - mean_m) / noiseless_var.sqrt()
        mvs_rescaled_mc = (normalized_mvs - rho * normalized_samples) / (1 - rho**2)
        cdf_mvs = normal.cdf(normalized_mvs).clamp_min(CLAMP_LB)
        cdf_rescaled_mvs = normal.cdf(mvs_rescaled_mc).clamp_min(CLAMP_LB)
        mv_ratio = cdf_rescaled_mvs / cdf_mvs

        log_term = torch.log(mv_ratio) + conditional_logprobs
        conditional_entropy = -(mv_ratio * log_term).mean(0)
        entropy_reduction = (
            initial_entropy - conditional_entropy.sum(dim=-2, keepdim=True)
        ).squeeze(-1)

        # average over the number of optima and squeeze the q-batch
        entropy_reduction = entropy_reduction.mean(dim=sample_dim).squeeze(-1)

        return entropy_reduction



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""Model fitting routines."""

from __future__ import annotations

from collections.abc import Callable, Sequence
from copy import deepcopy
from functools import partial
from itertools import filterfalse
from typing import Any
from warnings import catch_warnings, simplefilter, warn_explicit, WarningMessage

import torch

from botorch.exceptions.errors import ModelFittingError, UnsupportedError
from botorch.exceptions.warnings import OptimizationWarning
from botorch.logging import logger
from botorch.models import SingleTaskGP
from botorch.models.approximate_gp import ApproximateGPyTorchModel
from botorch.models.fully_bayesian import (
    AbstractFullyBayesianSingleTaskGP,
    SaasFullyBayesianSingleTaskGP,
)
from botorch.models.fully_bayesian_multitask import SaasFullyBayesianMultiTaskGP
from botorch.models.map_saas import get_map_saas_model
from botorch.models.model_list_gp_regression import ModelListGP
from botorch.models.transforms.input import InputTransform
from botorch.models.transforms.outcome import OutcomeTransform
from botorch.optim.closures import get_loss_closure_with_grads
from botorch.optim.core import _LBFGSB_MAXITER_MAXFUN_REGEX
from botorch.optim.fit import fit_gpytorch_mll_scipy, fit_gpytorch_mll_torch
from botorch.optim.utils import (
    _warning_handler_template,
    get_parameters,
    sample_all_priors,
)
from botorch.utils.context_managers import (
    module_rollback_ctx,
    parameter_rollback_ctx,
    TensorCheckpoint,
)
from botorch.utils.dispatcher import Dispatcher, type_bypassing_encoder
from gpytorch.likelihoods import Likelihood
from gpytorch.mlls._approximate_mll import _ApproximateMarginalLogLikelihood
from gpytorch.mlls.exact_marginal_log_likelihood import ExactMarginalLogLikelihood
from gpytorch.mlls.marginal_log_likelihood import MarginalLogLikelihood
from gpytorch.mlls.sum_marginal_log_likelihood import SumMarginalLogLikelihood
from linear_operator.utils.errors import NotPSDError
from pyro.infer.mcmc import MCMC, NUTS
from torch import device, Tensor
from torch.distributions import HalfCauchy
from torch.nn import Parameter
from torch.utils.data import DataLoader


def _debug_warn(w: WarningMessage) -> bool:
    if _LBFGSB_MAXITER_MAXFUN_REGEX.search(str(w.message)):
        return True
    # TODO: Better handle cases where warning handling logic
    # affects both debug and rethrow functions.
    return False


def _rethrow_warn(w: WarningMessage) -> bool:
    if not issubclass(w.category, OptimizationWarning):
        return True
    if "Optimization timed out after" in str(w.message):
        return True
    return False


DEFAULT_WARNING_HANDLER = partial(
    _warning_handler_template,
    debug=_debug_warn,
    rethrow=_rethrow_warn,
)
FitGPyTorchMLL = Dispatcher("fit_gpytorch_mll", encoder=type_bypassing_encoder)


def fit_gpytorch_mll(
    mll: MarginalLogLikelihood,
    closure: Callable[[], tuple[Tensor, Sequence[Tensor | None]]] | None = None,
    optimizer: Callable | None = None,
    closure_kwargs: dict[str, Any] | None = None,
    optimizer_kwargs: dict[str, Any] | None = None,
    **kwargs: Any,
) -> MarginalLogLikelihood:
    r"""Clearing house for fitting models passed as GPyTorch MarginalLogLikelihoods.

    Args:
        mll: A GPyTorch MarginalLogLikelihood instance.
        closure: Forward-backward closure for obtaining objective values and gradients.
            Responsible for setting parameters' `grad` attributes. If no closure is
            provided, one will be obtained by calling `get_loss_closure_with_grads`.
        optimizer: User specified optimization algorithm. When `optimizer is None`,
            this keyword argument is omitted when calling the dispatcher.
        closure_kwargs: Keyword arguments passed when calling `closure`.
        optimizer_kwargs: A dictionary of keyword arguments passed when
            calling `optimizer`.
        **kwargs: Keyword arguments passed down through the dispatcher to
            fit subroutines. Unexpected keywords are ignored.

    Returns:
        The `mll` instance. If fitting succeeded, then `mll` will be in evaluation mode,
        i.e. `mll.training == False`. Otherwise, `mll` will be in training mode.
    """
    if optimizer is not None:  # defer to per-method defaults
        kwargs["optimizer"] = optimizer

    return FitGPyTorchMLL(
        mll,
        type(mll.likelihood),
        type(mll.model),
        closure=closure,
        closure_kwargs=closure_kwargs,
        optimizer_kwargs=optimizer_kwargs,
        **kwargs,
    )


@FitGPyTorchMLL.register(MarginalLogLikelihood, object, object)
def _fit_fallback(
    mll: MarginalLogLikelihood,
    _: type[object],
    __: type[object],
    *,
    closure: Callable[[], tuple[Tensor, Sequence[Tensor | None]]] | None = None,
    optimizer: Callable = fit_gpytorch_mll_scipy,
    closure_kwargs: dict[str, Any] | None = None,
    optimizer_kwargs: dict[str, Any] | None = None,
    max_attempts: int = 5,
    pick_best_of_all_attempts: bool = False,
    warning_handler: Callable[[WarningMessage], bool] = DEFAULT_WARNING_HANDLER,
    caught_exception_types: tuple[type[BaseException], ...] = (NotPSDError,),
    **ignore: Any,
) -> MarginalLogLikelihood:
    r"""Generic fallback method for fitting Gaussian processes.

    Attempts to fit a model using the provided optimizer, then determines whether or
    not to retry by evaluating a given policy on emitted warning messages. The first
    attempt is run using the initialized parameter values; subsequent attempts begin
    by resampling tunable parameters.

    Args:
        closure: Forward-backward closure for obtaining objective values and gradients.
            Responsible for setting parameters' `grad` attributes. If no closure is
            provided, one will be obtained by calling `get_loss_closure_with_grads`.
        optimizer: The underlying optimization algorithm to run. Should return
            an `OptimizationResult` object, whose `fval` field records the negative
            MLL value. Defaults to `fit_gpytorch_mll_scipy`.
        closure_kwargs: Keyword arguments passed to `closure`.
        optimizer_kwargs: Keyword arguments passed to `optimizer`.
        max_attempts: The maximum number of fit attempts allowed. The attempt budget
            is NOT shared between calls to this method.
        pick_best_of_all_attempts: If True, the model will be fit `max_attempts` times,
            and the attempt that produces largest MLL value will be returned.
            First attempt uses the initial hyper parameter values, the subsequent
            attempts will call `sample_all_priors` to sample the initial values.
            If any attempt produces an error, the resulting parameters are discarded.
            If optimizer timeout is used, the `timeout_sec` will be used as is for
            each attempt, and it should be manually adjusted accordingly.
        warning_handler: A function used to filter warnings produced when calling
            `optimizer`. Any unfiltered warnings (those for which `warning_handler`
            returns `False`) will be rethrown and trigger a model fitting retry.
        caught_exception_types: A tuple of exception types whose instances should
            be logged at the `DEBUG` level.
        **ignore: This function ignores unrecognized keyword arguments.

    Returns:
        The `mll` instance. If fitting succeeded, then `mll` will be in evaluation mode,
        i.e. `mll.training == False`. Otherwise, `mll` will be in training mode.
    """
    # Setup
    optimizer_kwargs = {} if optimizer_kwargs is None else optimizer_kwargs
    params_nograd: dict[str, Parameter] = None  # pyre-ignore [9]
    ckpt_nograd: dict[str, TensorCheckpoint] = None  # pyre-ignore [9]
    ckpt: dict[str, TensorCheckpoint] = None  # pyre-ignore [9]

    # Build closure
    mll.train()
    if closure is None:
        closure = get_loss_closure_with_grads(
            mll, parameters=get_parameters(mll, requires_grad=True)
        )
    if closure_kwargs is not None:
        closure = partial(closure, **closure_kwargs)

    # Record best MLL & corresponding state dict.
    best_mll: float = -float("inf")
    best_state_dict = None
    # Attempt to fit the model
    for attempt in range(1, 1 + max_attempts):
        # Wrap with rollback contextmanager so that each loop iteration reloads the
        # original state_dict upon exiting (unless we clear `ckpt`).
        with module_rollback_ctx(mll, checkpoint=ckpt, device=device("cpu")) as ckpt:
            if attempt > 1:  # resample free parameters
                if params_nograd is None:
                    params_nograd = get_parameters(mll, requires_grad=False)

                if ckpt_nograd is None:  # reuse primary checkpoint
                    ckpt_nograd = {name: ckpt[name] for name in params_nograd}

                with parameter_rollback_ctx(params_nograd, checkpoint=ckpt_nograd):
                    sample_all_priors(mll.model)

            try:
                # Fit the model
                with catch_warnings(record=True) as warning_list:
                    simplefilter("always", category=OptimizationWarning)
                    result = optimizer(mll, closure=closure, **optimizer_kwargs)

                # Resolve warnings and determine whether or not to retry
                success = True
                for w in filterfalse(warning_handler, warning_list):
                    warn_explicit(str(w.message), w.category, w.filename, w.lineno)
                    success = False

                if success and not pick_best_of_all_attempts:
                    # If not picking best of all attempts, return the first
                    # successful attempt.
                    ckpt.clear()  # do not rollback upon exiting
                    return mll.eval()
                elif success:
                    # Update best MLL and corresponding state dict.
                    # Optimizers minimize negative MLL, so we negate fval.
                    current_mll = -result.fval
                    if current_mll > best_mll:
                        best_mll = current_mll
                        # Deepcopy is important here, otherwise they get updated.
                        best_state_dict = deepcopy(mll.state_dict())
                        message = f"Fit attempt #{attempt}: New best MLL: {best_mll}."
                    else:
                        message = (
                            f"Fit attempt #{attempt}: Current MLL {current_mll} did "
                            f"not beat best MLL so far {best_mll}."
                        )
                    logger.debug(message)

                # Ensure mll is in the right mode if going for another attempt.
                mll = mll if mll.training else mll.train()
                if not success:
                    logger.debug(
                        f"Fit attempt #{attempt} of {max_attempts} triggered retry "
                        f"policy {'.' if attempt == max_attempts else '; retrying...'}",
                    )

            except caught_exception_types as err:
                logger.debug(
                    f"Fit attempt #{attempt} of {max_attempts} failed with exception:\n"
                    f"{err}",
                )

    # If picking best of all attempts, return MLL with best state dict.
    if best_state_dict is not None:
        mll.load_state_dict(best_state_dict)
        return mll.eval()

    raise ModelFittingError("All attempts to fit the model have failed.")


@FitGPyTorchMLL.register(SumMarginalLogLikelihood, object, ModelListGP)
def _fit_list(
    mll: SumMarginalLogLikelihood,
    _: type[Likelihood],
    __: type[ModelListGP],
    **kwargs: Any,
) -> SumMarginalLogLikelihood:
    r"""Fitting routine for lists of independent Gaussian processes.

    Args:
        **kwargs: Passed to each of `mll.mlls`.

    Returns:
        The `mll` instance. If fitting succeeded for all of `mll.mlls`, then `mll` will
        be in evaluation mode, i.e. `mll.training == False`. Otherwise, `mll` will be in
        training mode.
    """
    mll.train()
    for sub_mll in mll.mlls:
        fit_gpytorch_mll(sub_mll, **kwargs)

    return mll.eval() if not any(sub_mll.training for sub_mll in mll.mlls) else mll


@FitGPyTorchMLL.register(_ApproximateMarginalLogLikelihood, object, object)
def _fit_fallback_approximate(
    mll: _ApproximateMarginalLogLikelihood,
    _: type[Likelihood],
    __: type[ApproximateGPyTorchModel],
    *,
    closure: Callable[[], tuple[Tensor, Sequence[Tensor | None]]] | None = None,
    data_loader: DataLoader | None = None,
    optimizer: Callable | None = None,
    full_batch_limit: int = 1024,
    **kwargs: Any,
) -> _ApproximateMarginalLogLikelihood:
    r"""Fallback method for fitting approximate Gaussian processes.

    Args:
        closure: Forward-backward closure for obtaining objective values and gradients.
            Responsible for setting parameters' `grad` attributes. If no closure is
            provided, one will be obtained by calling `get_loss_closure_with_grads`.
        optimizer: The underlying optimization algorithm to run. Default to
            `fit_gpytorch_mll_scipy` when `closure=None` and the model's internal
            training set has no more than `full_batch_cutoff` observations; otherwise,
            defaults to `fit_gpytorch_mll_torch`.
        data_loader: An optional DataLoader to pass to `get_loss_closure_with_grads`.
            May only be provided when `closure=None`.
        full_batch_limit: Threshold for determining the default choice of `optimizer`
            when `closure=None`.
        **kwargs: Keyword arguments passed to `_fit_fallback`.
    """
    if data_loader is not None:
        if closure is not None:
            raise UnsupportedError(
                "Only one of `data_loader` or `closure` may be passed."
            )
        closure = get_loss_closure_with_grads(
            mll=mll,
            data_loader=data_loader,
            parameters=get_parameters(mll, requires_grad=True),
        )

    if optimizer is None:
        optimizer = (
            fit_gpytorch_mll_scipy
            if closure is None and len(mll.model.train_targets) <= full_batch_limit
            else fit_gpytorch_mll_torch
        )

    return _fit_fallback(mll, _, __, closure=closure, optimizer=optimizer, **kwargs)


def fit_fully_bayesian_model_nuts(
    model: AbstractFullyBayesianSingleTaskGP | SaasFullyBayesianMultiTaskGP,
    max_tree_depth: int = 6,
    warmup_steps: int = 512,
    num_samples: int = 256,
    thinning: int = 16,
    disable_progbar: bool = False,
    jit_compile: bool = False,
) -> None:
    r"""Fit a fully Bayesian model using the No-U-Turn-Sampler (NUTS)


    Args:
        model: Fully Bayesian GP to be fitted.
        max_tree_depth: Maximum tree depth for NUTS
        warmup_steps: The number of burn-in steps for NUTS.
        num_samples:  The number of MCMC samples. Note that with thinning,
            num_samples / thinning samples are retained.
        thinning: The amount of thinning. Every nth sample is retained.
        disable_progbar: A boolean indicating whether to print the progress
            bar and diagnostics during MCMC.
        jit_compile: Whether to use jit. Using jit may be ~2X faster (rough estimate),
            but it will also increase the memory usage and sometimes result in runtime
            errors, e.g., https://github.com/pyro-ppl/pyro/issues/3136.

    Example:
        >>> gp = SaasFullyBayesianSingleTaskGP(train_X, train_Y)
        >>> fit_fully_bayesian_model_nuts(gp)
    """
    model.train()

    # Do inference with NUTS
    nuts = NUTS(
        model.pyro_model.sample,
        jit_compile=jit_compile,
        full_mass=True,
        ignore_jit_warnings=True,
        max_tree_depth=max_tree_depth,
    )
    mcmc = MCMC(
        nuts,
        warmup_steps=warmup_steps,
        num_samples=num_samples,
        disable_progbar=disable_progbar,
    )
    mcmc.run()

    # Get final MCMC samples from the Pyro model
    mcmc_samples = model.pyro_model.postprocess_mcmc_samples(
        mcmc_samples=mcmc.get_samples()
    )
    for k, v in mcmc_samples.items():
        mcmc_samples[k] = v[::thinning]

    # Load the MCMC samples back into the BoTorch model
    model.load_mcmc_samples(mcmc_samples)
    model.eval()


def get_fitted_map_saas_model(
    train_X: Tensor,
    train_Y: Tensor,
    train_Yvar: Tensor | None = None,
    input_transform: InputTransform | None = None,
    outcome_transform: OutcomeTransform | None = None,
    tau: float | None = None,
    optimizer_kwargs: dict[str, Any] | None = None,
) -> SingleTaskGP:
    """Get a fitted MAP SAAS model with a Matern kernel.

    Args:
        train_X: Tensor of shape `n x d` with training inputs.
        train_Y: Tensor of shape `n x 1` with training targets.
        train_Yvar: Optional tensor of shape `n x 1` with observed noise,
            inferred if None.
        input_transform: An optional input transform.
        outcome_transform: An optional outcome transforms.
        tau: Fixed value of the global shrinkage tau. If None, the model
            places a HC(0.1) prior on tau.
        optimizer_kwargs: A dict of options for the optimizer passed
            to fit_gpytorch_mll.

    Returns:
        A fitted SingleTaskGP with a Matern kernel.
    """
    # make sure optimizer_kwargs is a Dict
    optimizer_kwargs = optimizer_kwargs or {}
    model = get_map_saas_model(
        train_X=train_X,
        train_Y=train_Y,
        train_Yvar=train_Yvar,
        input_transform=(
            input_transform.train() if input_transform is not None else None
        ),
        outcome_transform=outcome_transform,
        tau=tau,
    )
    mll = ExactMarginalLogLikelihood(model=model, likelihood=model.likelihood)
    fit_gpytorch_mll(mll, optimizer_kwargs=optimizer_kwargs)
    return model


def get_fitted_map_saas_ensemble(
    train_X: Tensor,
    train_Y: Tensor,
    train_Yvar: Tensor | None = None,
    input_transform: InputTransform | None = None,
    outcome_transform: OutcomeTransform | None = None,
    taus: Tensor | list[float] | None = None,
    num_taus: int = 4,
    optimizer_kwargs: dict[str, Any] | None = None,
) -> SaasFullyBayesianSingleTaskGP:
    """Get a fitted SAAS ensemble using several different tau values.

    Args:
        train_X: Tensor of shape `n x d` with training inputs.
        train_Y: Tensor of shape `n x 1` with training targets.
        train_Yvar: Optional tensor of shape `n x 1` with observed noise,
            inferred if None.
        input_transform: An optional input transform.
        outcome_transform: An optional outcome transforms.
        taus: Global shrinkage values to use. If None, we sample `num_taus` values
            from an HC(0.1) distrbution.
        num_taus: Optional argument for how many taus to sample.
        optimizer_kwargs: A dict of options for the optimizer passed
            to fit_gpytorch_mll.

    Returns:
        A fitted SaasFullyBayesianSingleTaskGP with a Matern kernel.
    """
    tkwargs = {"device": train_X.device, "dtype": train_X.dtype}
    if taus is None:
        taus = HalfCauchy(0.1).sample([num_taus]).to(**tkwargs)
    num_samples = len(taus)
    if num_samples == 1:
        raise ValueError(
            "Use `get_fitted_map_saas_model` if you only specify one value of tau"
        )

    mean = torch.zeros(num_samples, **tkwargs)
    outputscale = torch.zeros(num_samples, **tkwargs)
    lengthscale = torch.zeros(num_samples, train_X.shape[-1], **tkwargs)
    noise = torch.zeros(num_samples, **tkwargs)

    # Fit a model for each tau and save the hyperparameters
    for i, tau in enumerate(taus):
        model = get_fitted_map_saas_model(
            train_X,
            train_Y,
            train_Yvar=train_Yvar,
            input_transform=input_transform,
            outcome_transform=outcome_transform,
            tau=tau,
            optimizer_kwargs=optimizer_kwargs,
        )
        mean[i] = model.mean_module.constant.detach().clone()
        outputscale[i] = model.covar_module.outputscale.detach().clone()
        lengthscale[i, :] = model.covar_module.base_kernel.lengthscale.detach().clone()
        if train_Yvar is None:
            noise[i] = model.likelihood.noise.detach().clone()

    # Load the samples into a fully Bayesian SAAS model
    ensemble_model = SaasFullyBayesianSingleTaskGP(
        train_X=train_X,
        train_Y=train_Y,
        train_Yvar=train_Yvar,
        input_transform=(
            input_transform.train() if input_transform is not None else None
        ),
        outcome_transform=outcome_transform,
    )
    mcmc_samples = {
        "mean": mean,
        "outputscale": outputscale,
        "lengthscale": lengthscale,
    }
    if train_Yvar is None:
        mcmc_samples["noise"] = noise
    ensemble_model.train()
    ensemble_model.load_mcmc_samples(mcmc_samples=mcmc_samples)
    ensemble_model.eval()
    return ensemble_model



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Utilities for speeding up optimization in tests.

"""

from __future__ import annotations

from collections.abc import Generator
from contextlib import contextmanager, ExitStack
from functools import wraps
from typing import Any, Callable
from unittest import mock

from botorch.optim.batched_lbfgs_b import fmin_l_bfgs_b_batched

from botorch.optim.initializers import (
    gen_batch_initial_conditions,
    gen_one_shot_kg_initial_conditions,
)
from botorch.optim.utils.timeout import minimize_with_timeout
from scipy.optimize import OptimizeResult
from torch import Tensor


@contextmanager
def mock_optimize_context_manager(
    force: bool = False,
) -> Generator[None, None, None]:
    """A context manager that uses mocks to speed up optimization for testing.
    Currently, the primary tactic is to force the underlying scipy methods to stop
    after just one iteration.

        force: If True will not raise an AssertionError if no mocks are called.
            USE RESPONSIBLY.
    """

    def two_iteration_minimize(*args: Any, **kwargs: Any) -> OptimizeResult:
        if kwargs["options"] is None:
            kwargs["options"] = {}
        # Using two iterations here to allow SLSQP to adapt to constraints.
        kwargs["options"]["maxiter"] = 2
        return minimize_with_timeout(*args, **kwargs)

    def two_iteration_fast_minimize(*args: Any, **kwargs: Any) -> OptimizeResult:
        # Using two iterations here to allow SLSQP to adapt to constraints.
        kwargs["maxiter"] = 2
        return fmin_l_bfgs_b_batched(*args, **kwargs)

    def minimal_gen_ics(*args: Any, **kwargs: Any) -> Tensor:
        kwargs["num_restarts"] = 2
        kwargs["raw_samples"] = 4

        return gen_batch_initial_conditions(*args, **kwargs)

    def minimal_gen_os_ics(*args: Any, **kwargs: Any) -> Tensor | None:
        kwargs["num_restarts"] = 2
        kwargs["raw_samples"] = 4

        return gen_one_shot_kg_initial_conditions(*args, **kwargs)

    with ExitStack() as es:
        # Note this `minimize_with_timeout` is defined in optim.utils.timeout;
        # this mock only has an effect when calling a function used in
        # `botorch.generation.gen`, such as `gen_candidates_scipy`.
        mock_generation = es.enter_context(
            mock.patch(
                "botorch.generation.gen.minimize_with_timeout",
                wraps=two_iteration_minimize,
            )
        )

        mock_generation_fast = es.enter_context(
            mock.patch(
                "botorch.generation.gen.fmin_l_bfgs_b_batched",
                wraps=two_iteration_fast_minimize,
            )
        )

        # Similarly, works when using calling a function defined in
        # `optim.core`, such as `scipy_minimize` and `torch_minimize`.
        mock_fit = es.enter_context(
            mock.patch(
                "botorch.optim.core.minimize_with_timeout",
                wraps=two_iteration_minimize,
            )
        )

        # Works when calling a function in `optim.optimize` such as
        # `optimize_acqf`
        mock_gen_ics = es.enter_context(
            mock.patch(
                "botorch.optim.optimize.gen_batch_initial_conditions",
                wraps=minimal_gen_ics,
            )
        )

        # Works when calling a function in `optim.optimize` such as
        # `optimize_acqf`
        mock_gen_os_ics = es.enter_context(
            mock.patch(
                "botorch.optim.optimize.gen_one_shot_kg_initial_conditions",
                wraps=minimal_gen_os_ics,
            )
        )

        # Reduce default number of iterations in `optimize_acqf_mixed_alternating`.
        for name in [
            "MAX_ITER_ALTER",
            "MAX_ITER_DISCRETE",
            "MAX_ITER_CONT",
        ]:
            es.enter_context(mock.patch(f"botorch.optim.optimize_mixed.{name}", new=1))

        yield

    if (not force) and all(
        mock_.call_count < 1
        for mock_ in [
            mock_generation,
            mock_generation_fast,
            mock_fit,
            mock_gen_ics,
            mock_gen_os_ics,
        ]
    ):
        raise AssertionError(
            "No mocks were called in the context manager. Please remove unused "
            "mock_optimize_context_manager()."
        )


def mock_optimize(f: Callable) -> Callable:
    """Wraps `f` in `mock_optimize_context_manager` for use as a decorator."""

    @wraps(f)
    # pyre-fixme[3]: Return type must be annotated.
    def inner(*args: Any, **kwargs: Any):
        with mock_optimize_context_manager():
            return f(*args, **kwargs)

    return inner



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Helpers for computing statistical distances between Gaussian distributions.

Contributor: hvarfner
"""

import torch
from torch import Tensor


def mvn_kl_divergence(
    p_mean: Tensor, q_mean: Tensor, p_covar: Tensor, q_covar: Tensor
) -> Tensor:
    """Computes the KL-divergence between two multivariate normal distributions.
    Args:
        p_mean: A `batch_shape x dist_shape x 1`-dim Tensor of means of
            the first distribution.
        q_mean: A `batch_shape x dist_shape x dist_shape`-dim Tensor of
            covariances of the first distribution,  where the covariances
            are (optionally) in the q-batch dim.
        p_mean: A `batch_shape x dist_shape x 1`-dim Tensor of means of
            the second distribution.
        q_mean: A `batch_shape x dist_shape x dist_shape`-dim Tensor of
            covariances of the second distribution where the covariances
            are (optionally) in the q-batch dim.
    Returns:
        A tensor of shape `batch_shape x dist_shape` denoting the KL-divergence
        between the multivariate Gaussian distributions p and q.
    """
    p_inv_covar = torch.inverse(p_covar)
    mean_diff = p_mean - q_mean
    kl_first_term = torch.diagonal(
        torch.matmul(p_inv_covar, q_covar), dim1=-2, dim2=-1
    ).sum(-1, keepdim=True)
    kl_second_term = torch.matmul(
        torch.matmul(mean_diff.transpose(-2, -1), p_inv_covar), mean_diff
    ).squeeze(-1)
    kl_third_term = (torch.logdet(p_covar) - torch.logdet(q_covar)).unsqueeze(-1)
    return 0.5 * (kl_first_term + kl_second_term + kl_third_term - p_mean.shape[-2])


def mvn_hellinger_distance(
    p_mean: Tensor, q_mean: Tensor, p_covar: Tensor, q_covar: Tensor
) -> Tensor:
    """Computes the (2)-Hellinger distance between two multivariate normal
        distributions.
    Args:
        p_mean: A `batch_shape x dist_shape x 1`-dim Tensor of means of
            the first distribution.
        q_mean: A `batch_shape x dist_shape x dist_shape`-dim Tensor of
            covariances of the first distribution,  where the covariances
            are (optionally) in the q-batch dim.
        p_mean: A `batch_shape x dist_shape x 1`-dim Tensor of means of
            the second distribution.
        q_mean: A `batch_shape x dist_shape x dist_shape`-dim Tensor of
            covariances of the second distribution where the covariances
            are (optionally) in the q-batch dim.
    Returns:
        A tensor of shape `batch_shape x dist_shape` denoting the KL-divergence
        between the multivariate Gaussian distributions p and q.
    """
    p_logdet = torch.logdet(p_covar).unsqueeze(-1)
    q_logdet = torch.logdet(q_covar).unsqueeze(-1)
    avg_covar = (p_covar + q_covar) / 2

    # we need to re-use the cholesky decomp, so we compute it once here
    L_avg = torch.linalg.cholesky(avg_covar)
    L_inv = torch.inverse(L_avg)

    # removes one dimension, which needs to be recouped
    pq_logdet = (
        torch.pow(torch.diagonal(L_avg, dim1=-2, dim2=-1), 2)
        .prod(dim=-1, keepdim=True)
        .log()
    )
    base_logterm = 0.25 * (p_logdet + q_logdet) - 0.5 * pq_logdet

    mean_diff = p_mean - q_mean
    L_mean_diff = torch.matmul(L_inv, mean_diff)
    exp_logterm = -0.125 * torch.matmul(L_mean_diff.transpose(-2, -1), L_mean_diff)
    sq_hdist = 1 - (base_logterm + exp_logterm.squeeze(-1)).exp()
    return sq_hdist.sqrt()



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
Acquisition function for Self-Correcting Bayesian Optimization [hvarfner2023scorebo]_.

References

.. [hvarfner2023scorebo]
    C. Hvarfner, E. Hellsten, F. Hutter, L. Nardi.
    Self-Correcting Bayesian Optimization thorugh Bayesian Active Learning.
    In Proceedings of the Annual Conference on Neural Information
    Processing Systems (NeurIPS), 2023.

Contributor: hvarfner
"""

from __future__ import annotations

import warnings
from typing import Optional

import torch
from botorch import settings
from botorch.acquisition.acquisition import MCSamplerMixin
from botorch.acquisition.bayesian_active_learning import (
    FullyBayesianAcquisitionFunction,
)
from botorch.acquisition.objective import ScalarizedPosteriorTransform
from botorch.models.fully_bayesian import MCMC_DIM, SaasFullyBayesianSingleTaskGP
from botorch.models.utils import fantasize as fantasize_flag
from botorch.models.utils.gpytorch_modules import MIN_INFERRED_NOISE_LEVEL
from botorch.utils.transforms import (
    average_over_ensemble_models,
    concatenate_pending_points,
    t_batch_mode_transform,
)
from botorch_community.acquisition.bayesian_active_learning import DISTANCE_METRICS
from torch import Tensor

# The lower bound on the CDF value of the max-values
CLAMP_LB = 1e-6


class qSelfCorrectingBayesianOptimization(
    FullyBayesianAcquisitionFunction, MCSamplerMixin
):
    def __init__(
        self,
        model: SaasFullyBayesianSingleTaskGP,
        optimal_outputs: Tensor,
        optimal_inputs: Optional[Tensor] = None,
        X_pending: Optional[Tensor] = None,
        distance_metric: Optional[str] = "hellinger",
        posterior_transform: Optional[ScalarizedPosteriorTransform] = None,
    ) -> None:
        r"""Self-correcting Bayesian optimization [hvarfner2023scorebo]_ acquisition
        function. SCoreBO seeks to find accurate hyperparameters during the course
        of optimization by incorporating an active learning-like objective into
        optimization.

        Args:
            model: A fully bayesian model single-outcome model.
            optimal_inputs: A `num_samples x num_models x d`-dim tensor containing
                the sampled optimal inputs of dimension `d`.
            optimal_outputs: A `num_samples x num_models x 1`-dim Tensor containing
                the optimal objective values.
            X_pending: A `batch_shape, m x d`-dim Tensor of `m` design points
            distance_metric: The distance metric used. Defaults to
                "hellinger".
        """

        super().__init__(model=model)
        # To enable fully bayesian GP conditioning, we need to unsqueeze
        # to get num_optima x num_gps unique GPs
        self.optimal_outputs = optimal_outputs.unsqueeze(-2)
        self.optimal_output_values = (
            posterior_transform.evaluate(self.optimal_outputs).unsqueeze(-1)
            if posterior_transform
            else self.optimal_outputs
        )
        self.posterior_transform = posterior_transform

        # JES-like version of SCoreBO if optimal inputs are provided
        if optimal_inputs is not None:
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore")
                with fantasize_flag():
                    with settings.propagate_grads(False):
                        # We must do a forward pass one before conditioning.
                        self.model.posterior(
                            self.model.train_inputs[0], observation_noise=False
                        )

            self.optimal_inputs = optimal_inputs.unsqueeze(-2)
            self.conditional_model = self.model.condition_on_observations(
                X=self.model.transform_inputs(self.optimal_inputs),
                Y=self.optimal_outputs,
                noise=torch.full_like(self.optimal_outputs, MIN_INFERRED_NOISE_LEVEL),
            )
        # otherwise, we do a MES-like variant (which places vastly more emphasis on
        # HP learning as supposed to optimization)
        else:
            self.conditional_model = self.model

        self.normal = torch.distributions.Normal(
            torch.zeros(
                1, device=self.optimal_outputs.device, dtype=self.optimal_outputs.dtype
            ),
            torch.ones(
                1, device=self.optimal_outputs.device, dtype=self.optimal_outputs.dtype
            ),
        )
        if distance_metric not in DISTANCE_METRICS.keys():
            raise ValueError(  # pragma: no cover
                f"Distance metric need to be one of {list(DISTANCE_METRICS.keys())}"
            )
        self.distance = DISTANCE_METRICS[distance_metric]
        self.set_X_pending(X_pending)

    @concatenate_pending_points
    @t_batch_mode_transform()
    @average_over_ensemble_models
    def forward(self, X: Tensor) -> Tensor:
        # since we have two MC dims (over models and optima), we need to
        # unsqueeze a second dim to accomodate the posterior pass
        prev_posterior = self.model.posterior(
            X.unsqueeze(MCMC_DIM),
            observation_noise=True,
            posterior_transform=self.posterior_transform,
        )
        noiseless_posterior = self.conditional_model.posterior(
            X.unsqueeze(MCMC_DIM),
            observation_noise=False,
            posterior_transform=self.posterior_transform,
        )
        posterior = self.conditional_model.posterior(
            X.unsqueeze(MCMC_DIM),
            observation_noise=True,
            posterior_transform=self.posterior_transform,
        )

        marg_mean = prev_posterior.mean.mean(dim=MCMC_DIM, keepdim=True)
        cond_means = posterior.mean
        cond_variances = posterior.variance
        cond_covar = posterior.covariance_matrix

        # the mixture variance is squeezed, need it unsqueezed
        marg_covar = prev_posterior.mixture_covariance_matrix.unsqueeze(MCMC_DIM)
        noiseless_var = noiseless_posterior.variance
        normalized_mvs = (
            self.optimal_output_values - cond_means
        ) / noiseless_var.sqrt()
        cdf_mvs = self.normal.cdf(normalized_mvs).clamp_min(CLAMP_LB)
        pdf_mvs = torch.exp(self.normal.log_prob(normalized_mvs))
        mean_truncated = cond_means - noiseless_var.sqrt() * pdf_mvs / cdf_mvs

        # This is the noiseless variance (i.e. the part that gets truncated)
        var_truncated = noiseless_var * (
            1 - normalized_mvs * pdf_mvs / cdf_mvs - torch.pow(pdf_mvs / cdf_mvs, 2)
        )
        # and add the (possibly heteroskedastic) noise
        var_truncated = var_truncated + (cond_variances - noiseless_var)

        # truncating the entire covariance matrix is not trivial, so we assume the
        # truncation is proportional on the off-diags as on the diagonals and scale
        # the covariance accordingly for all elements in the q-batch (if there is one)
        # for q=1, this is equivalent to simply truncating the posterior
        covar_scaling = (var_truncated / cond_variances).sqrt()
        trunc_covar = covar_scaling.transpose(-1, -2) * covar_scaling * cond_covar
        dist = self.distance(mean_truncated, marg_mean, trunc_covar, marg_covar)

        # squeeze output dim and average over optimal samples dim (MCMC_DIM).
        # Model dim is averaged later
        return dist.mean(MCMC_DIM).sum(-1)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from botorch.acquisition.acquisition import (
    AcquisitionFunction,
    OneShotAcquisitionFunction,
)
from botorch.acquisition.active_learning import (
    PairwiseMCPosteriorVariance,
    qNegIntegratedPosteriorVariance,
)
from botorch.acquisition.analytic import (
    AnalyticAcquisitionFunction,
    ConstrainedExpectedImprovement,
    ExpectedImprovement,
    LogExpectedImprovement,
    LogNoisyExpectedImprovement,
    NoisyExpectedImprovement,
    PosteriorMean,
    PosteriorStandardDeviation,
    ProbabilityOfImprovement,
    qAnalyticProbabilityOfImprovement,
    UpperConfidenceBound,
)
from botorch.acquisition.bayesian_active_learning import (
    qBayesianActiveLearningByDisagreement,
)
from botorch.acquisition.cost_aware import (
    GenericCostAwareUtility,
    InverseCostWeightedUtility,
)
from botorch.acquisition.decoupled import DecoupledAcquisitionFunction
from botorch.acquisition.factory import get_acquisition_function
from botorch.acquisition.fixed_feature import FixedFeatureAcquisitionFunction
from botorch.acquisition.input_constructors import get_acqf_input_constructor
from botorch.acquisition.knowledge_gradient import (
    qKnowledgeGradient,
    qMultiFidelityKnowledgeGradient,
)
from botorch.acquisition.logei import (
    LogImprovementMCAcquisitionFunction,
    qLogExpectedImprovement,
    qLogNoisyExpectedImprovement,
)
from botorch.acquisition.max_value_entropy_search import (
    MaxValueBase,
    qLowerBoundMaxValueEntropy,
    qMaxValueEntropy,
    qMultiFidelityLowerBoundMaxValueEntropy,
    qMultiFidelityMaxValueEntropy,
)
from botorch.acquisition.monte_carlo import (
    MCAcquisitionFunction,
    qExpectedImprovement,
    qLowerConfidenceBound,
    qNoisyExpectedImprovement,
    qPosteriorStandardDeviation,
    qProbabilityOfImprovement,
    qSimpleRegret,
    qUpperConfidenceBound,
    SampleReducingMCAcquisitionFunction,
)
from botorch.acquisition.multi_step_lookahead import qMultiStepLookahead
from botorch.acquisition.multioutput_acquisition import MultiOutputAcquisitionFunction
from botorch.acquisition.objective import (
    ConstrainedMCObjective,
    GenericMCObjective,
    IdentityMCObjective,
    LearnedObjective,
    LinearMCObjective,
    MCAcquisitionObjective,
    ScalarizedPosteriorTransform,
)
from botorch.acquisition.preference import (
    AnalyticExpectedUtilityOfBestOption,
    PairwiseBayesianActiveLearningByDisagreement,
    qExpectedUtilityOfBestOption,
)
from botorch.acquisition.prior_guided import PriorGuidedAcquisitionFunction
from botorch.acquisition.proximal import ProximalAcquisitionFunction

__all__ = [
    "AcquisitionFunction",
    "AnalyticAcquisitionFunction",
    "AnalyticExpectedUtilityOfBestOption",
    "ConstrainedExpectedImprovement",
    "DecoupledAcquisitionFunction",
    "ExpectedImprovement",
    "LogExpectedImprovement",
    "LogNoisyExpectedImprovement",
    "FixedFeatureAcquisitionFunction",
    "GenericCostAwareUtility",
    "InverseCostWeightedUtility",
    "NoisyExpectedImprovement",
    "OneShotAcquisitionFunction",
    "PairwiseBayesianActiveLearningByDisagreement",
    "PairwiseMCPosteriorVariance",
    "PosteriorMean",
    "PosteriorStandardDeviation",
    "PriorGuidedAcquisitionFunction",
    "ProbabilityOfImprovement",
    "ProximalAcquisitionFunction",
    "UpperConfidenceBound",
    "qBayesianActiveLearningByDisagreement",
    "qAnalyticProbabilityOfImprovement",
    "qExpectedImprovement",
    "qExpectedUtilityOfBestOption",
    "LogImprovementMCAcquisitionFunction",
    "qLogExpectedImprovement",
    "qLogNoisyExpectedImprovement",
    "qKnowledgeGradient",
    "MaxValueBase",
    "qMultiFidelityKnowledgeGradient",
    "qMaxValueEntropy",
    "qMultiFidelityLowerBoundMaxValueEntropy",
    "qLowerBoundMaxValueEntropy",
    "qMultiFidelityMaxValueEntropy",
    "qMultiStepLookahead",
    "qNoisyExpectedImprovement",
    "qNegIntegratedPosteriorVariance",
    "qProbabilityOfImprovement",
    "qSimpleRegret",
    "qPosteriorStandardDeviation",
    "qLowerConfidenceBound",
    "qUpperConfidenceBound",
    "ConstrainedMCObjective",
    "GenericMCObjective",
    "IdentityMCObjective",
    "LearnedObjective",
    "LinearMCObjective",
    "MCAcquisitionFunction",
    "SampleReducingMCAcquisitionFunction",
    "MCAcquisitionObjective",
    "ScalarizedPosteriorTransform",
    "get_acquisition_function",
    "get_acqf_input_constructor",
    "MultiOutputAcquisitionFunction",
]



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Optional, Union

import torch
from botorch.acquisition.objective import (
    ExpectationPosteriorTransform,
    PosteriorTransform,
    ScalarizedPosteriorTransform,
)
from botorch.exceptions.errors import UnsupportedError
from botorch.utils.testing import BotorchTestCase
from botorch_community.acquisition.discretized import (
    DiscretizedExpectedImprovement,
    DiscretizedProbabilityOfImprovement,
)
from botorch_community.posteriors.riemann import BoundedRiemannPosterior
from torch import Tensor


class MockDiscretizedModel:
    def __init__(self, borders, probabilities):
        """Mock model for testing discretized acquisition functions.

        This class simulates a model that returns a BoundedRiemannPosterior with
        predefined borders and probabilities, regardless of the input X.

        Attributes:
            borders: A tensor representing the boundaries of the discretized bins.
            probabilities: A tensor representing the probability mass for each bin.
        """

        self.borders = borders
        self.probabilities = probabilities

    def posterior(
        self,
        X: Tensor,
        output_indices: Optional[list[int]] = None,
        observation_noise: Union[bool, Tensor] = False,
        posterior_transform: Optional[PosteriorTransform] = None,
    ) -> BoundedRiemannPosterior:
        return BoundedRiemannPosterior(self.borders, self.probabilities)


class TestDiscretizedExpectedImprovement(BotorchTestCase):
    def test_ag_integrate(self):
        for dtype in (torch.float, torch.double):
            tkwargs = {"device": self.device, "dtype": dtype}
            acqf = DiscretizedExpectedImprovement(None, best_f=2.0)

            # best_f <= lower
            lower_bound = torch.tensor(1.0, **tkwargs)
            upper_bound = torch.tensor(1.5, **tkwargs)
            improvement = acqf.ag_integrate(lower_bound, upper_bound)
            self.assertEqual(improvement, 0.0)

            # lower < best_f <= upper
            lower_bound = torch.tensor(1.0)
            upper_bound = torch.tensor(3.0)
            improvement = acqf.ag_integrate(lower_bound, upper_bound)
            self.assertEqual(improvement, (2.0 + 3.0) / 2 - 2.0)

            # upper <= best_f
            lower_bound = torch.tensor(3.0)
            upper_bound = torch.tensor(5.0)
            improvement = acqf.ag_integrate(lower_bound, upper_bound)
            self.assertEqual(improvement, ((5.0 + 3.0) / 2 - 2.0) * 2.0)

    def test_ag_integrate_batch(self):
        for dtype in (torch.float, torch.double):
            tkwargs = {"device": self.device, "dtype": dtype}
            acqf = DiscretizedExpectedImprovement(
                None, best_f=torch.tensor([[2.0], [3.0]], **tkwargs)
            )

            lower_bound = torch.tensor([[1.0], [2.0]], **tkwargs)
            upper_bound = torch.tensor([[1.5], [6.0]], **tkwargs)
            improvement = acqf.ag_integrate(lower_bound, upper_bound)
            self.assertEqual(improvement[0], 0.0)
            self.assertEqual(improvement[1], 3.0**2 / 2)

    def test_forward(self):
        for dtype in (torch.double, torch.float):
            tkwargs = {"device": self.device, "dtype": dtype}

            borders = torch.tensor([0.0, 1.0, 3.0, 6.0], **tkwargs)
            probs = torch.tensor([0.2, 0.3, 0.5], **tkwargs)
            mock_model = MockDiscretizedModel(borders, probs)

            acqf = DiscretizedExpectedImprovement(mock_model, best_f=2.0)
            # value will not be used since posterior is mocked
            X = torch.tensor([[1.0]])
            acqf_values = acqf(X)

            # EI is 0 for bucket 1, (1.0**2)/2 = 0.5 for bucket 2,
            # and 2.5*3. = 7.5 for bucket 3
            # so the expected value is
            # 0.2 / 1. * 0 + 0.3 / 2. * 0.5 + 0.5 / 3. * 7.5 = 1.325
            self.assertAlmostEqual(acqf_values.item(), 1.325)

    def test_forward_approximating_normal(self):
        tkwargs = {"device": self.device, "dtype": torch.double}
        borders = torch.linspace(-8, 8, 10_000, **tkwargs)

        mean = 1.1
        std = 1.3
        target_dist = torch.distributions.Normal(mean, std)
        probs = target_dist.cdf(borders[1:]) - target_dist.cdf(borders[:-1])

        mock_model = MockDiscretizedModel(borders, probs)
        # Test that the discretized EI approximates the analytical EI
        # for a normal distribution
        best_f = 0.5
        acqf = DiscretizedExpectedImprovement(mock_model, best_f=best_f)
        X = torch.tensor([[1.0]])  # value will not be used since posterior is mocked
        acqf_values = acqf(X)

        # Calculate analytical EI for normal distribution
        improvement = mean - best_f
        z = torch.tensor(improvement / std)
        normal_dist = torch.distributions.Normal(0, 1)
        analytical_ei = (
            improvement * normal_dist.cdf(z) + std * normal_dist.log_prob(z).exp()
        )

        # Check that the discretized EI is close to the analytical EI
        self.assertLess(torch.abs(acqf_values - analytical_ei) / analytical_ei, 0.01)

        # minimization task
        acqf = DiscretizedExpectedImprovement(
            mock_model,
            best_f=-best_f,
            posterior_transform=ScalarizedPosteriorTransform(
                weights=torch.tensor([-1.0])
            ),
        )
        acqf_values = acqf(X)

        # Calculate analytical EI for normal distribution
        improvement = -mean + best_f
        z = torch.tensor(improvement / std)
        normal_dist = torch.distributions.Normal(0, 1)
        analytical_ei = (
            improvement * normal_dist.cdf(z) + std * normal_dist.log_prob(z).exp()
        )

        # Check that the discretized EI is close to the analytical EI
        self.assertLess(torch.abs(acqf_values - analytical_ei) / analytical_ei, 0.01)

    def test_unsupported_posterior_transform(self):
        """Test that UnsupportedError is raised for non-ScalarizedPosteriorTransform."""
        mock_model = MockDiscretizedModel(
            torch.tensor([0.0, 1.0, 2.0]), torch.tensor([0.5, 0.5])
        )

        # Test with ExpectationPosteriorTransform
        expectation_transform = ExpectationPosteriorTransform(n_w=2)
        with self.assertRaises(UnsupportedError):
            DiscretizedExpectedImprovement(
                mock_model, best_f=1.0, posterior_transform=expectation_transform
            )


class TestDiscretizedProbabilityofImprovement(BotorchTestCase):
    def test_ag_integrate(self):
        for dtype in (torch.float, torch.double):
            tkwargs = {"device": self.device, "dtype": dtype}
            acqf = DiscretizedProbabilityOfImprovement(None, best_f=2.0)

            # best_f <= lower
            lower_bound = torch.tensor(1.0, **tkwargs)
            upper_bound = torch.tensor(1.5, **tkwargs)
            improvement = acqf.ag_integrate(lower_bound, upper_bound)
            self.assertEqual(improvement, 0.0)

            # lower < best_f <= upper
            lower_bound = torch.tensor(1.0)
            upper_bound = torch.tensor(3.0)
            improvement = acqf.ag_integrate(lower_bound, upper_bound)
            self.assertEqual(improvement, 1.0)

            # upper <= best_f
            lower_bound = torch.tensor(3.0)
            upper_bound = torch.tensor(5.0)
            improvement = acqf.ag_integrate(lower_bound, upper_bound)
            self.assertEqual(improvement, 2.0)

    def test_ag_integrate_batch(self):
        for dtype in (torch.float, torch.double):
            tkwargs = {"device": self.device, "dtype": dtype}
            acqf = DiscretizedProbabilityOfImprovement(
                None, best_f=torch.tensor([[2.0], [3.0]], **tkwargs)
            )

            lower_bound = torch.tensor([[1.0], [2.0]], **tkwargs)
            upper_bound = torch.tensor([[1.5], [6.0]], **tkwargs)
            improvement = acqf.ag_integrate(lower_bound, upper_bound)
            self.assertEqual(improvement[0], 0.0)
            self.assertEqual(improvement[1], 3.0)

    def test_forward(self):
        borders = torch.tensor([0.0, 1.0, 3.0, 6.0])
        probs = torch.tensor([0.2, 0.3, 0.5])
        mock_model = MockDiscretizedModel(borders, probs)

        acqf = DiscretizedProbabilityOfImprovement(mock_model, best_f=2.0)
        X = torch.tensor([[1.0]])  # value will not be used since posterior is mocked
        acqf_values = acqf(X)

        # PI is 0 for bucket 1, 0.5 for bucket 2, and 1 for bucket 3
        # so the final probability is 0.2 * 0 + 0.3 * 0.5 + 0.5 * 1 = 1.4
        self.assertAlmostEqual(acqf_values.item(), 0.65)

    def test_forward_approximating_normal(self):
        tkwargs = {"device": self.device, "dtype": torch.double}
        borders = torch.linspace(-8, 8, 10_000, **tkwargs)

        mean = 1.1
        std = 1.3
        target_dist = torch.distributions.Normal(mean, std)
        probs = target_dist.cdf(borders[1:]) - target_dist.cdf(borders[:-1])

        mock_model = MockDiscretizedModel(borders, probs)

        # Compute PI
        best_f = 0.5
        acqf = DiscretizedProbabilityOfImprovement(mock_model, best_f=best_f)
        X = torch.tensor([[1.0]])  # value will not be used since posterior is mocked
        acqf_values = acqf(X)

        # Calculate analytical PI for normal distribution
        prob_improvement = 1.0 - target_dist.cdf(torch.tensor(best_f))

        # Check that the discretized EI is close to the analytical EI
        self.assertLess(
            torch.abs(acqf_values - prob_improvement) / prob_improvement, 0.01
        )

        # minimization task
        acqf = DiscretizedProbabilityOfImprovement(
            mock_model,
            best_f=-best_f,
            posterior_transform=ScalarizedPosteriorTransform(
                weights=torch.tensor([-1.0])
            ),
        )
        acqf_values = acqf(X)

        # Calculate analytical PI for normal distribution
        prob_improvement = target_dist.cdf(torch.tensor(best_f))

        # Check that the discretized EI is close to the analytical EI
        self.assertLess(
            torch.abs(acqf_values - prob_improvement) / prob_improvement, 0.01
        )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
Abstract model class for all GPyTorch-based botorch models.

To implement your own, simply inherit from both the provided classes and a
GPyTorch Model class such as an ExactGP.
"""

from __future__ import annotations

import itertools
import warnings
from abc import ABC
from copy import deepcopy
from typing import Any, TYPE_CHECKING

import torch
from botorch.acquisition.objective import PosteriorTransform
from botorch.exceptions.errors import (
    BotorchTensorDimensionError,
    InputDataError,
    UnsupportedError,
)
from botorch.exceptions.warnings import (
    _get_single_precision_warning,
    BotorchTensorDimensionWarning,
    InputDataWarning,
)
from botorch.models.model import Model, ModelList
from botorch.models.utils import (
    _make_X_full,
    add_output_dim,
    gpt_posterior_settings,
    mod_batch_shape,
    multioutput_to_batch_mode_transform,
)
from botorch.models.utils.assorted import fantasize as fantasize_flag
from botorch.posteriors.fully_bayesian import GaussianMixturePosterior
from botorch.posteriors.gpytorch import GPyTorchPosterior
from botorch.utils.multitask import separate_mtmvn
from botorch.utils.transforms import is_ensemble
from gpytorch.distributions import MultitaskMultivariateNormal, MultivariateNormal
from gpytorch.likelihoods.gaussian_likelihood import FixedNoiseGaussianLikelihood
from linear_operator.operators import BlockDiagLinearOperator, CatLinearOperator
from torch import broadcast_shapes, Tensor

if TYPE_CHECKING:
    from botorch.posteriors.posterior_list import PosteriorList  # pragma: no cover
    from botorch.posteriors.transformed import TransformedPosterior  # pragma: no cover
    from gpytorch.likelihoods import Likelihood  # pragma: no cover


class GPyTorchModel(Model, ABC):
    r"""Abstract base class for models based on GPyTorch models.

    The easiest way to use this is to subclass a model from a GPyTorch model
    class (e.g. an `ExactGP`) and this `GPyTorchModel`. See e.g. `SingleTaskGP`.
    """

    likelihood: Likelihood

    @staticmethod
    def _validate_tensor_args(
        X: Tensor, Y: Tensor, Yvar: Tensor | None = None, strict: bool = True
    ) -> None:
        r"""Checks that `Y` and `Yvar` have an explicit output dimension if strict.
        Checks that the dtypes of the inputs match, and warns if using float.

        This also checks that `Yvar` has the same trailing dimensions as `Y`. Note
        we only infer that an explicit output dimension exists when `X` and `Y` have
        the same `batch_shape`.

        Args:
            X: A `batch_shape x n x d`-dim Tensor, where `d` is the dimension of
                the feature space, `n` is the number of points per batch, and
                `batch_shape` is the batch shape (potentially empty).
            Y: A `batch_shape' x n x m`-dim Tensor, where `m` is the number of
                model outputs, `n'` is the number of points per batch, and
                `batch_shape'` is the batch shape of the observations.
            Yvar: A `batch_shape' x n x m` tensor of observed measurement noise.
                Note: this will be None when using a model that infers the noise
                level (e.g. a `SingleTaskGP`).
            strict: A boolean indicating whether to check that `Y` and `Yvar`
                have an explicit output dimension.
        """
        if X.dim() != Y.dim():
            if (X.dim() - Y.dim() == 1) and (X.shape[:-1] == Y.shape):
                message = (
                    "An explicit output dimension is required for targets."
                    f" Expected Y with dimension {X.dim()} (got {Y.dim()=})."
                )
            else:
                message = (
                    "Expected X and Y to have the same number of dimensions"
                    f" (got X with dimension {X.dim()} and Y with dimension"
                    f" {Y.dim()})."
                )
            if strict:
                raise BotorchTensorDimensionError(message)
            else:
                warnings.warn(
                    "Non-strict enforcement of botorch tensor conventions. The "
                    "following error would have been raised with strict enforcement: "
                    f"{message}",
                    BotorchTensorDimensionWarning,
                    stacklevel=2,
                )
        # Yvar may not have the same batch dimensions, but the trailing dimensions
        # of Yvar should be the same as the trailing dimensions of Y.
        if Yvar is not None and Y.shape[-(Yvar.dim()) :] != Yvar.shape:
            raise BotorchTensorDimensionError(
                "An explicit output dimension is required for observation noise."
                f" Expected Yvar with shape: {Y.shape[-Yvar.dim() :]} (got"
                f" {Yvar.shape})."
            )
        # Check the dtypes.
        if X.dtype != Y.dtype or (Yvar is not None and Y.dtype != Yvar.dtype):
            raise InputDataError(
                "Expected all inputs to share the same dtype. Got "
                f"{X.dtype} for X, {Y.dtype} for Y, and "
                f"{Yvar.dtype if Yvar is not None else None} for Yvar."
            )
        if X.dtype != torch.float64:
            warnings.warn(
                _get_single_precision_warning(str(X.dtype)),
                InputDataWarning,
                stacklevel=3,  # Warn at model constructor call.
            )

    @property
    def batch_shape(self) -> torch.Size:
        r"""The batch shape of the model.

        This is a batch shape from an I/O perspective, independent of the internal
        representation of the model (as e.g. in BatchedMultiOutputGPyTorchModel).
        For a model with `m` outputs, a `test_batch_shape x q x d`-shaped input `X`
        to the `posterior` method returns a Posterior object over an output of
        shape `broadcast(test_batch_shape, model.batch_shape) x q x m`.
        """
        return self.train_inputs[0].shape[:-2]

    @property
    def num_outputs(self) -> int:
        r"""The number of outputs of the model."""
        return self._num_outputs

    # pyre-fixme[14]: Inconsistent override.
    # `botorch.models.gpytorch.GPyTorchModel.posterior` overrides method defined
    # in `Model` inconsistently. Could not find parameter `output_indices` in
    # overriding signature.
    def posterior(
        self,
        X: Tensor,
        observation_noise: bool | Tensor = False,
        posterior_transform: PosteriorTransform | None = None,
        **kwargs: Any,
    ) -> GPyTorchPosterior | TransformedPosterior:
        r"""Computes the posterior over model outputs at the provided points.

        Args:
            X: A `(batch_shape) x q x d`-dim Tensor, where `d` is the dimension
                of the feature space and `q` is the number of points considered
                jointly.
            observation_noise: If True, add the observation noise from the
                likelihood to the posterior. If a Tensor, use it directly as the
                observation noise (must be of shape `(batch_shape) x q`). It is
                assumed to be in the outcome-transformed space if an outcome
                transform is used.
            posterior_transform: An optional PosteriorTransform.

        Returns:
            A `GPyTorchPosterior` object, representing a batch of `b` joint
            distributions over `q` points. Includes observation noise if
            specified.
        """
        self.eval()  # make sure model is in eval mode
        # input transforms are applied at `posterior` in `eval` mode, and at
        # `model.forward()` at the training time
        X = self.transform_inputs(X)
        with gpt_posterior_settings():
            # NOTE: BoTorch's GPyTorchModels also inherit from GPyTorch's ExactGP, thus
            # self(X) calls GPyTorch's ExactGP's __call__, which computes the posterior,
            # rather than e.g. SingleTaskGP's forward, which computes the prior.
            mvn = self(X)
            if observation_noise is not False:
                if isinstance(observation_noise, torch.Tensor):
                    # TODO: Make sure observation noise is transformed correctly
                    self._validate_tensor_args(X=X, Y=observation_noise)
                    if observation_noise.size(-1) == 1:
                        observation_noise = observation_noise.squeeze(-1)
                    mvn = self.likelihood(mvn, X, noise=observation_noise)
                else:
                    mvn = self.likelihood(mvn, X)
        posterior = GPyTorchPosterior(distribution=mvn)
        if hasattr(self, "outcome_transform"):
            posterior = self.outcome_transform.untransform_posterior(posterior, X=X)
        if posterior_transform is not None:
            return posterior_transform(posterior)
        return posterior

    def condition_on_observations(
        self, X: Tensor, Y: Tensor, noise: Tensor | None = None, **kwargs: Any
    ) -> Model:
        r"""Condition the model on new observations.

        Args:
            X: A `batch_shape x n' x d`-dim Tensor, where `d` is the dimension of
                the feature space, `n'` is the number of points per batch, and
                `batch_shape` is the batch shape (must be compatible with the
                batch shape of the model).
            Y: A `batch_shape' x n x m`-dim Tensor, where `m` is the number of
                model outputs, `n'` is the number of points per batch, and
                `batch_shape'` is the batch shape of the observations.
                `batch_shape'` must be broadcastable to `batch_shape` using
                standard broadcasting semantics. If `Y` has fewer batch dimensions
                than `X`, its is assumed that the missing batch dimensions are
                the same for all `Y`.
            noise: If not `None`, a tensor of the same shape as `Y` representing
                the associated noise variance.
            kwargs: Passed to `self.get_fantasy_model`.

        Returns:
            A `Model` object of the same type, representing the original model
            conditioned on the new observations `(X, Y)` (and possibly noise
            observations passed in via kwargs).

        Example:
            >>> train_X = torch.rand(20, 2)
            >>> train_Y = torch.sin(train_X[:, 0]) + torch.cos(train_X[:, 1])
            >>> model = SingleTaskGP(train_X, train_Y)
            >>> new_X = torch.rand(5, 2)
            >>> new_Y = torch.sin(new_X[:, 0]) + torch.cos(new_X[:, 1])
            >>> model = model.condition_on_observations(X=new_X, Y=new_Y)
        """
        Yvar = noise

        if hasattr(self, "outcome_transform"):
            # pass the transformed data to get_fantasy_model below
            # (unless we've already trasnformed if BatchedMultiOutputGPyTorchModel)
            if not isinstance(self, BatchedMultiOutputGPyTorchModel):
                # `noise` is assumed to already be outcome-transformed.
                Y, _ = self.outcome_transform(Y=Y, Yvar=Yvar, X=X)
        # Validate using strict=False, since we cannot tell if Y has an explicit
        # output dimension. Do not check shapes when fantasizing as they are
        # not expected to match.
        if fantasize_flag.off():
            self._validate_tensor_args(X=X, Y=Y, Yvar=Yvar, strict=False)
        if Y.size(-1) == 1:
            Y = Y.squeeze(-1)
            if Yvar is not None:
                kwargs.update({"noise": Yvar.squeeze(-1)})
        # get_fantasy_model will properly copy any existing outcome transforms
        # (since it deepcopies the original model)

        return self.get_fantasy_model(inputs=X, targets=Y, **kwargs)


# pyre-fixme[13]: uninitialized attributes _num_outputs, _input_batch_shape,
# _aug_batch_shape
class BatchedMultiOutputGPyTorchModel(GPyTorchModel):
    r"""Base class for batched multi-output GPyTorch models with independent outputs.

    This model should be used when the same training data is used for all outputs.
    Outputs are modeled independently by using a different batch for each output.
    """

    _num_outputs: int
    _input_batch_shape: torch.Size
    _aug_batch_shape: torch.Size

    @staticmethod
    def get_batch_dimensions(
        train_X: Tensor, train_Y: Tensor
    ) -> tuple[torch.Size, torch.Size]:
        r"""Get the raw batch shape and output-augmented batch shape of the inputs.

        Args:
            train_X: A `n x d` or `batch_shape x n x d` (batch mode) tensor of training
                features.
            train_Y: A `n x m` or `batch_shape x n x m` (batch mode) tensor of
                training observations.

        Returns:
            2-element tuple containing

            - The `input_batch_shape`
            - The output-augmented batch shape: `input_batch_shape x (m)`
        """
        input_batch_shape = train_X.shape[:-2]
        aug_batch_shape = input_batch_shape
        num_outputs = train_Y.shape[-1]
        if num_outputs > 1:
            aug_batch_shape += torch.Size([num_outputs])
        return input_batch_shape, aug_batch_shape

    def _set_dimensions(self, train_X: Tensor, train_Y: Tensor) -> None:
        r"""Store the number of outputs and the batch shape.

        Args:
            train_X: A `n x d` or `batch_shape x n x d` (batch mode) tensor of training
                features.
            train_Y: A `n x m` or `batch_shape x n x m` (batch mode) tensor of
                training observations.
        """
        self._num_outputs = train_Y.shape[-1]
        self._input_batch_shape, self._aug_batch_shape = self.get_batch_dimensions(
            train_X=train_X, train_Y=train_Y
        )

    @property
    def batch_shape(self) -> torch.Size:
        r"""The batch shape of the model.

        This is a batch shape from an I/O perspective, independent of the internal
        representation of the model (as e.g. in BatchedMultiOutputGPyTorchModel).
        For a model with `m` outputs, a `test_batch_shape x q x d`-shaped input `X`
        to the `posterior` method returns a Posterior object over an output of
        shape `broadcast(test_batch_shape, model.batch_shape) x q x m`.
        """
        return self._input_batch_shape

    def _transform_tensor_args(
        self, X: Tensor, Y: Tensor, Yvar: Tensor | None = None
    ) -> tuple[Tensor, Tensor, Tensor | None]:
        r"""Transforms tensor arguments: for single output models, the output
        dimension is squeezed and for multi-output models, the output dimension is
        transformed into the left-most batch dimension.

        Args:
            X: A `n x d` or `batch_shape x n x d` (batch mode) tensor of training
                features.
            Y: A `n x m` or `batch_shape x n x m` (batch mode) tensor of
                training observations.
            Yvar: A `n x m` or `batch_shape x n x m` (batch mode) tensor of
                observed measurement noise. Note: this will be None when using a model
                that infers the noise level (e.g. a `SingleTaskGP`).

        Returns:
            3-element tuple containing

            - A `input_batch_shape x (m) x n x d` tensor of training features.
            - A `target_batch_shape x (m) x n` tensor of training observations.
            - A `target_batch_shape x (m) x n` tensor observed measurement noise
                (or None).
        """
        if self._num_outputs > 1:
            return multioutput_to_batch_mode_transform(
                train_X=X, train_Y=Y, train_Yvar=Yvar, num_outputs=self._num_outputs
            )
        return X, Y.squeeze(-1), None if Yvar is None else Yvar.squeeze(-1)

    def _apply_noise(
        self,
        X: Tensor,
        mvn: MultivariateNormal,
        observation_noise: bool | Tensor = False,
    ) -> MultivariateNormal:
        """Adds the observation noise to the posterior.

        Args:
            X: A tensor of shape `batch_shape x q x d`.
            mvn: A `MultivariateNormal` object representing the posterior over the true
                latent function.
            num_outputs: The number of outputs of the model.
            observation_noise: If True, add the observation noise from the
                likelihood to the posterior. If a Tensor, use it directly as the
                observation noise (must be of shape `(batch_shape) x q x m`).

        Returns:
            The posterior predictive.
        """
        if observation_noise is False:
            return mvn
        # noise_shape is `broadcast(test_batch_shape, model.batch_shape) x m x q`
        noise_shape = mvn.batch_shape + mvn.event_shape
        if torch.is_tensor(observation_noise):
            # TODO: Validate noise shape
            # make observation_noise's shape match noise_shape
            if self.num_outputs > 1:
                obs_noise = observation_noise.transpose(-1, -2)
            else:
                obs_noise = observation_noise.squeeze(-1)
            mvn = self.likelihood(
                mvn,
                X,
                noise=obs_noise.expand(noise_shape),
            )
        elif isinstance(self.likelihood, FixedNoiseGaussianLikelihood):
            # Use the mean of the previous noise values (TODO: be smarter here).
            observation_noise = self.likelihood.noise.mean(dim=-1, keepdim=True)
            mvn = self.likelihood(
                mvn,
                X,
                noise=observation_noise.expand(noise_shape),
            )
        else:
            mvn = self.likelihood(mvn, X)
        return mvn

    # pyre-ignore[14]: Inconsistent override. Could not find parameter
    # `Keywords(typing.Any)` in overriding signature.
    def posterior(
        self,
        X: Tensor,
        output_indices: list[int] | None = None,
        observation_noise: bool | Tensor = False,
        posterior_transform: PosteriorTransform | None = None,
    ) -> GPyTorchPosterior | TransformedPosterior:
        r"""Computes the posterior over model outputs at the provided points.

        Args:
            X: A `(batch_shape) x q x d`-dim Tensor, where `d` is the dimension
                of the feature space and `q` is the number of points considered
                jointly.
            output_indices: A list of indices, corresponding to the outputs over
                which to compute the posterior (if the model is multi-output).
                Can be used to speed up computation if only a subset of the
                model's outputs are required for optimization. If omitted,
                computes the posterior over all model outputs.
            observation_noise: If True, add the observation noise from the
                likelihood to the posterior. If a Tensor, use it directly as the
                observation noise (must be of shape `(batch_shape) x q x m`).
            posterior_transform: An optional PosteriorTransform.

        Returns:
            A `GPyTorchPosterior` object, representing `batch_shape` joint
            distributions over `q` points and the outputs selected by
            `output_indices` each. Includes observation noise if specified.
        """
        self.eval()  # make sure model is in eval mode
        # input transforms are applied at `posterior` in `eval` mode, and at
        # `model.forward()` at the training time
        X = self.transform_inputs(X)
        with gpt_posterior_settings():
            # insert a dimension for the output dimension
            if self._num_outputs > 1:
                X, output_dim_idx = add_output_dim(
                    X=X, original_batch_shape=self._input_batch_shape
                )
            # NOTE: BoTorch's GPyTorchModels also inherit from GPyTorch's ExactGP, thus
            # self(X) calls GPyTorch's ExactGP's __call__, which computes the posterior,
            # rather than e.g. SingleTaskGP's forward, which computes the prior.
            mvn = self(X)
            mvn = self._apply_noise(X=X, mvn=mvn, observation_noise=observation_noise)
            if self._num_outputs > 1:
                if torch.jit.is_tracing():
                    mvn = MultitaskMultivariateNormal.from_batch_mvn(
                        mvn, task_dim=output_dim_idx
                    )
                else:
                    mean_x = mvn.mean
                    covar_x = mvn.lazy_covariance_matrix
                    output_indices = output_indices or range(self._num_outputs)
                    mvns = [
                        MultivariateNormal(
                            mean_x.select(dim=output_dim_idx, index=t),
                            covar_x[(slice(None),) * output_dim_idx + (t,)],
                        )
                        for t in output_indices
                    ]
                    mvn = MultitaskMultivariateNormal.from_independent_mvns(mvns=mvns)

        posterior = GPyTorchPosterior(distribution=mvn)
        if hasattr(self, "outcome_transform"):
            posterior = self.outcome_transform.untransform_posterior(posterior, X=X)
        if posterior_transform is not None:
            return posterior_transform(posterior)
        return posterior

    # pyre-ignore[14]: Inconsistent override. Could not find parameter `noise`.
    def condition_on_observations(
        self, X: Tensor, Y: Tensor, **kwargs: Any
    ) -> BatchedMultiOutputGPyTorchModel:
        r"""Condition the model on new observations.

        Args:
            X: A `batch_shape x n' x d`-dim Tensor, where `d` is the dimension of
                the feature space, `m` is the number of points per batch, and
                `batch_shape` is the batch shape (must be compatible with the
                batch shape of the model).
            Y: A `batch_shape' x n' x m`-dim Tensor, where `m` is the number of
                model outputs, `n'` is the number of points per batch, and
                `batch_shape'` is the batch shape of the observations.
                `batch_shape'` must be broadcastable to `batch_shape` using
                standard broadcasting semantics. If `Y` has fewer batch dimensions
                than `X`, its is assumed that the missing batch dimensions are
                the same for all `Y`.

        Returns:
            A `BatchedMultiOutputGPyTorchModel` object of the same type with
            `n + n'` training examples, representing the original model
            conditioned on the new observations `(X, Y)` (and possibly noise
            observations passed in via kwargs).

        Example:
            >>> train_X = torch.rand(20, 2)
            >>> train_Y = torch.cat(
            >>>     [torch.sin(train_X[:, 0]), torch.cos(train_X[:, 1])], -1
            >>> )
            >>> model = SingleTaskGP(train_X, train_Y)
            >>> new_X = torch.rand(5, 2)
            >>> new_Y = torch.cat([torch.sin(new_X[:, 0]), torch.cos(new_X[:, 1])], -1)
            >>> model = model.condition_on_observations(X=new_X, Y=new_Y)
        """
        noise = kwargs.get("noise")
        if hasattr(self, "outcome_transform"):
            # We need to apply transforms before shifting batch indices around.
            # `noise` is assumed to already be outcome-transformed.
            Y, _ = self.outcome_transform(Y, X=X)
        # Do not check shapes when fantasizing as they are not expected to match.
        if fantasize_flag.off():
            self._validate_tensor_args(X=X, Y=Y, Yvar=noise, strict=False)
        inputs = X
        if self._num_outputs > 1:
            inputs, targets, noise = multioutput_to_batch_mode_transform(
                train_X=X, train_Y=Y, num_outputs=self._num_outputs, train_Yvar=noise
            )
            # `multioutput_to_batch_mode_transform` removes the output dimension,
            # which is necessary for `condition_on_observations`
            targets = targets.unsqueeze(-1)
            if noise is not None:
                noise = noise.unsqueeze(-1)
        else:
            inputs = X
            targets = Y
        if noise is not None:
            kwargs.update({"noise": noise})
        fantasy_model = super().condition_on_observations(X=inputs, Y=targets, **kwargs)
        fantasy_model._input_batch_shape = fantasy_model.train_targets.shape[
            : (-1 if self._num_outputs == 1 else -2)
        ]
        if not self._is_fully_bayesian:
            fantasy_model._aug_batch_shape = fantasy_model.train_targets.shape[:-1]
        return fantasy_model

    def subset_output(self, idcs: list[int]) -> BatchedMultiOutputGPyTorchModel:
        r"""Subset the model along the output dimension.

        Args:
            idcs: The output indices to subset the model to.

        Returns:
            The current model, subset to the specified output indices.
        """
        try:
            subset_batch_dict = self._subset_batch_dict
        except AttributeError:
            raise NotImplementedError(
                "`subset_output` requires the model to define a `_subset_batch_dict` "
                "attribute that lists the indices of the output dimensions in each "
                "model parameter that needs to be subset."
            )

        m = len(idcs)
        new_model = deepcopy(self)

        subset_everything = self.num_outputs == m and idcs == list(range(m))
        if subset_everything:
            return new_model

        tidxr = torch.tensor(idcs, device=new_model.train_targets.device)
        idxr = tidxr if m > 1 else idcs[0]
        new_tail_bs = torch.Size([m]) if m > 1 else torch.Size()

        new_model._num_outputs = m
        new_model._aug_batch_shape = new_model._aug_batch_shape[:-1] + new_tail_bs
        new_model.train_inputs = tuple(
            ti[..., idxr, :, :] for ti in new_model.train_inputs
        )
        new_model.train_targets = new_model.train_targets[..., idxr, :]

        # adjust batch shapes of parameters/buffers if necessary
        for full_name, p in itertools.chain(
            new_model.named_parameters(), new_model.named_buffers()
        ):
            if full_name in subset_batch_dict:
                idx = subset_batch_dict[full_name]
                new_data = p.index_select(dim=idx, index=tidxr)
                if m == 1:
                    new_data = new_data.squeeze(idx)
                p.data = new_data
            mod_name = full_name.split(".")[:-1]
            mod_batch_shape(new_model, mod_name, m if m > 1 else 0)

        # subset outcome transform if present
        try:
            subset_octf = new_model.outcome_transform.subset_output(idcs=idcs)
            new_model.outcome_transform = subset_octf
        except AttributeError:
            pass

        # Subset fixed noise likelihood if present.
        if isinstance(self.likelihood, FixedNoiseGaussianLikelihood):
            full_noise = new_model.likelihood.noise_covar.noise
            new_noise = full_noise[..., idcs if len(idcs) > 1 else idcs[0], :]
            new_model.likelihood.noise_covar.noise = new_noise

        return new_model


class ModelListGPyTorchModel(ModelList, GPyTorchModel, ABC):
    r"""Abstract base class for models based on multi-output GPyTorch models.

    This is meant to be used with a gpytorch ModelList wrapper for independent
    evaluation of submodels. Those submodels can themselves be multi-output
    models, in which case the task covariances will be ignored.
    """

    @property
    def batch_shape(self) -> torch.Size:
        r"""The batch shape of the model.

        This is a batch shape from an I/O perspective, independent of the internal
        representation of the model (as e.g. in BatchedMultiOutputGPyTorchModel).
        For a model with `m` outputs, a `test_batch_shape x q x d`-shaped input `X`
        to the `posterior` method returns a Posterior object over an output of
        shape `broadcast(test_batch_shape, model.batch_shape) x q x m`.
        """
        batch_shapes = {m.batch_shape for m in self.models}
        if len(batch_shapes) > 1:
            msg = (
                f"Component models of {self.__class__.__name__} have different "
                "batch shapes"
            )
            try:
                broadcast_shape = torch.broadcast_shapes(*batch_shapes)
                warnings.warn(msg + ". Broadcasting batch shapes.", stacklevel=2)
                return broadcast_shape
            except RuntimeError:
                raise NotImplementedError(msg + " that are not broadcastble.")
        return next(iter(batch_shapes))

    # pyre-fixme[14]: Inconsistent override in return types
    def posterior(
        self,
        X: Tensor,
        output_indices: list[int] | None = None,
        observation_noise: bool | Tensor = False,
        posterior_transform: PosteriorTransform | None = None,
    ) -> GPyTorchPosterior | PosteriorList:
        r"""Computes the posterior over model outputs at the provided points.
        If any model returns a MultitaskMultivariateNormal posterior, then that
        will be split into individual MVNs per task, with inter-task covariance
        ignored.

        Args:
            X: A `b x q x d`-dim Tensor, where `d` is the dimension of the
                feature space, `q` is the number of points considered jointly,
                and `b` is the batch dimension.
            output_indices: A list of indices, corresponding to the outputs over
                which to compute the posterior (if the model is multi-output).
                Can be used to speed up computation if only a subset of the
                model's outputs are required for optimization. If omitted,
                computes the posterior over all model outputs.
            observation_noise: If True, add the observation noise from the
                respective likelihoods to the posterior. If a Tensor of shape
                `(batch_shape) x q x m`, use it directly as the observation
                noise (with `observation_noise[...,i]` added to the posterior
                of the `i`-th model).
            posterior_transform: An optional PosteriorTransform.

        Returns:
            - If no `posterior_transform` is provided and the component models have no
                `outcome_transform`, or if the component models only use linear outcome
                transforms like `Standardize` (i.e. not `Log`), returns a
                `GPyTorchPosterior` or `GaussianMixturePosterior` object,
                representing `batch_shape` joint distributions over `q` points
                and the outputs selected by `output_indices` each. Includes
                measurement noise if `observation_noise` is specified.
            - If no `posterior_transform` is provided and component models have
                nonlinear transforms like `Log`, returns a `PosteriorList` with
                sub-posteriors of type `TransformedPosterior`
            - If `posterior_transform` is provided, that posterior transform will be
               applied and will determine the return type. This could potentially be
               any subclass of `Posterior`, but common choices give a
               `GPyTorchPosterior`.
        """

        # Nonlinear transforms untransform to a `TransformedPosterior`,
        # which can't be made into a `GPyTorchPosterior`
        returns_untransformed = any(
            hasattr(mod, "outcome_transform") and (not mod.outcome_transform._is_linear)
            for mod in self.models
        )
        # NOTE: We're not passing in the posterior transform here. We'll apply it later.
        posterior = ModelList.posterior(
            self,
            X=X,
            output_indices=output_indices,
            observation_noise=observation_noise,
        )
        if not returns_untransformed:
            mvns = [p.distribution for p in posterior.posteriors]
            if any(isinstance(m, MultitaskMultivariateNormal) for m in mvns):
                mvn_list = []
                for mvn in mvns:
                    if len(mvn.event_shape) == 2:
                        # We separate MTMVNs into independent-across-task MVNs for
                        # the convenience of using BlockDiagLinearOperator below.
                        # (b x q x m x m) -> list of m (b x q x 1 x 1)
                        mvn_list.extend(separate_mtmvn(mvn))
                    else:
                        mvn_list.append(mvn)
                mean = torch.stack([mvn.mean for mvn in mvn_list], dim=-1)
                covars = CatLinearOperator(
                    *[mvn.lazy_covariance_matrix.unsqueeze(-3) for mvn in mvn_list],
                    dim=-3,
                )  # List of m (b x q x 1 x 1) -> (b x q x m x 1 x 1)
                mvn = MultitaskMultivariateNormal(
                    mean=mean,
                    covariance_matrix=BlockDiagLinearOperator(covars, block_dim=-3).to(
                        X
                    ),  # (b x q x m x 1 x 1) -> (b x q x m x m)
                    interleaved=False,
                )
            else:
                mvns = self._broadcast_mvns(mvns=mvns)
                mvn = (
                    mvns[0]
                    if len(mvns) == 1
                    else MultitaskMultivariateNormal.from_independent_mvns(mvns=mvns)
                )
            # Return the result as a GPyTorchPosterior/GaussianMixturePosterior.
            if any(is_ensemble(m) for m in self.models):
                # Mixing fully Bayesian and other GP models is currently
                # not supported.
                posterior = GaussianMixturePosterior(distribution=mvn)
            else:
                posterior = GPyTorchPosterior(distribution=mvn)
        if posterior_transform is not None:
            return posterior_transform(posterior)
        return posterior

    def condition_on_observations(self, X: Tensor, Y: Tensor, **kwargs: Any) -> Model:
        raise NotImplementedError()

    def _broadcast_mvns(self, mvns: list[MultivariateNormal]) -> MultivariateNormal:
        """Broadcasts the batch shapes of the given MultivariateNormals.

        The MVNs will have a batch shape of `input_batch_shape x model_batch_shape`.
        If the model batch shapes are broadcastable, we will broadcast the mvns to
        a batch shape of `input_batch_shape x self.batch_shape`.

        Args:
            mvns: A list of MultivariateNormals.

        Returns:
            A list of MultivariateNormals with broadcasted batch shapes.
        """
        mvn_batch_shapes = {mvn.batch_shape for mvn in mvns}
        if len(mvn_batch_shapes) == 1:
            # All MVNs have the same batch shape. We can return as is.
            return mvns
        # This call will error out if they're not broadcastable.
        # If they're broadcastable, it'll log a warning.
        target_model_shape = self.batch_shape
        max_batch = max(mvn_batch_shapes, key=len)
        max_len = len(max_batch)
        input_batch_len = max_len - len(target_model_shape)
        for i in range(len(mvns)):  # Loop over index since we modify contents.
            while len(mvns[i].batch_shape) < max_len:
                # MVN is missing batch dimensions. Unsqueeze as needed.
                mvns[i] = mvns[i].unsqueeze(input_batch_len)
            if mvns[i].batch_shape != max_batch:
                # Expand to match the batch shapes.
                mvns[i] = mvns[i].expand(max_batch)
        return mvns


class MultiTaskGPyTorchModel(GPyTorchModel, ABC):
    r"""Abstract base class for multi-task models based on GPyTorch models.

    This class provides the `posterior` method to models that implement a
    "long-format" multi-task GP in the style of `MultiTaskGP`.
    """

    def _map_tasks(self, task_values: Tensor) -> Tensor:
        """Map raw task values to the task indices used by the model.

        Args:
            task_values: A tensor of task values.

        Returns:
            A tensor of task indices with the same shape as the input
                tensor.
        """
        if self._task_mapper is None:
            if not (
                torch.all(0 <= task_values) and torch.all(task_values < self.num_tasks)
            ):
                raise ValueError(
                    "Expected all task features in `X` to be between 0 and "
                    f"self.num_tasks - 1. Got {task_values}."
                )
        else:
            task_values = task_values.long()

            unexpected_task_values = set(task_values.unique().tolist()).difference(
                self._expected_task_values
            )
            if len(unexpected_task_values) > 0:
                raise ValueError(
                    "Received invalid raw task values. Expected raw value to be in"
                    f" {self._expected_task_values}, but got unexpected task values:"
                    f" {unexpected_task_values}."
                )
            task_values = self._task_mapper[task_values]
        return task_values

    def _apply_noise(
        self,
        X: Tensor,
        mvn: MultivariateNormal,
        observation_noise: bool | Tensor,
    ) -> MultivariateNormal:
        """Adds the observation noise to the posterior.

        If the likelihood is a `FixedNoiseGaussianLikelihood`, then
        the average noise per task is computed, and a diagonal noise
        matrix is added to the posterior covariance matrix, where
        the noise per input is the average noise for its respective
        task. If the likelihood is a Gaussian likelihood, then
        currently there is a shared inferred noise level for all
        tasks.

        TODO: implement support for task-specific inferred noise levels.

        Args:
            X: A tensor of shape `batch_shape x q x d + 1`,
                where `d` is the dimension of the feature space and the `+ 1`
                dimension is the task feature / index.
            mvn: A `MultivariateNormal` object representing the posterior over the true
                latent function.
            num_outputs: The number of outputs of the model.
            observation_noise: If True, add observation noise from the respective
                likelihood. Tensor input is currently not supported.

        Returns:
            The posterior predictive.
        """
        if torch.is_tensor(observation_noise):
            raise NotImplementedError(
                "Passing a tensor of observations is not supported by MultiTaskGP."
            )
        elif observation_noise is False:
            return mvn
        elif isinstance(self.likelihood, FixedNoiseGaussianLikelihood):
            # get task features for test points
            test_task_features = X[..., self._task_feature]
            test_task_features = self._map_tasks(test_task_features).long()
            unique_test_task_features = test_task_features.unique()
            # get task features for training points
            train_task_features = self.train_inputs[0][..., self._task_feature]
            train_task_features = self._map_tasks(train_task_features).long()
            noise_by_task = torch.zeros(
                *self.batch_shape, self.num_tasks, dtype=X.dtype, device=X.device
            )
            for task_feature in unique_test_task_features:
                mask = train_task_features == task_feature
                noise_by_task[..., task_feature] = self.likelihood.noise[
                    ..., mask
                ].mean(dim=-1)
            # noise_shape is `broadcast(test_batch_shape, model.batch_shape) x q`
            noise_shape = (
                broadcast_shapes(X.shape[:-2], self.batch_shape) + X.shape[-2:-1]
            )
            # Expand and gather ensures we pick correct noise dimensions for
            # batch evaluations of batched models.
            observation_noise = noise_by_task.expand(*noise_shape[:-1], -1).gather(
                dim=-1, index=test_task_features.expand(noise_shape)
            )
            return self.likelihood(
                mvn,
                X,
                noise=observation_noise,
            )
        return self.likelihood(mvn, X)

    # pyre-ignore[14]: Inconsistent override. Could not find parameter
    # `Keywords(typing.Any)` in overriding signature.
    def posterior(
        self,
        X: Tensor,
        output_indices: list[int] | None = None,
        observation_noise: bool | Tensor = False,
        posterior_transform: PosteriorTransform | None = None,
    ) -> GPyTorchPosterior | TransformedPosterior:
        r"""Computes the posterior over model outputs at the provided points.

        Args:
            X: A tensor of shape `batch_shape x q x d` or `batch_shape x q x (d + 1)`,
                where `d` is the dimension of the feature space (not including task
                indices) and `q` is the number of points considered jointly. The `+ 1`
                dimension is the optional task feature / index. If given, the model
                produces the outputs for the given task indices. If omitted, the
                model produces outputs for tasks in in `self._output_tasks` (specified
                as `output_tasks` while constructing the model), which can overwritten
                using `output_indices`.
            output_indices: A list of task values over which to compute the posterior.
                Only used if `X` does not include the task feature. If omitted,
                defaults to `self._output_tasks`.
            observation_noise: If True, add observation noise from the respective
                likelihoods. If a Tensor, specifies the observation noise levels
                to add.
            posterior_transform: An optional PosteriorTransform.

        Returns:
            A `GPyTorchPosterior` object, representing `batch_shape` joint
            distributions over `q` points. If the task features are included in `X`,
            the posterior will be single output. Otherwise, the posterior will be
            single or multi output corresponding to the tasks included in
            either the `output_indices` or `self._output_tasks`.
        """
        includes_task_feature = X.shape[-1] == self.num_non_task_features + 1
        if includes_task_feature:
            if output_indices is not None:
                raise ValueError(
                    "`output_indices` must be None when `X` includes task features."
                )
            task_features = X[..., self._task_feature].unique()
            num_outputs = 1
            X_full = X
        else:
            # Add the task features to construct the full X for evaluation.
            task_features = torch.tensor(
                self._output_tasks if output_indices is None else output_indices,
                dtype=torch.long,
                device=X.device,
            )
            num_outputs = len(task_features)
            X_full = _make_X_full(
                X=X, output_indices=task_features.tolist(), tf=self._task_feature
            )
        # Make sure all task feature values are valid.
        task_features = self._map_tasks(task_values=task_features)
        self.eval()  # make sure model is in eval mode
        # input transforms are applied at `posterior` in `eval` mode, and at
        # `model.forward()` at the training time
        X_full = self.transform_inputs(X_full)
        with gpt_posterior_settings():
            mvn = self(X_full)
            mvn = self._apply_noise(
                X=X_full,
                mvn=mvn,
                observation_noise=observation_noise,
            )
        # If single-output, return the posterior of a single-output model
        if num_outputs == 1:
            posterior = GPyTorchPosterior(distribution=mvn)
        else:
            # Otherwise, make a MultitaskMultivariateNormal out of this
            mtmvn = MultitaskMultivariateNormal(
                mean=mvn.mean.view(*mvn.mean.shape[:-1], num_outputs, -1).transpose(
                    -1, -2
                ),
                covariance_matrix=mvn.lazy_covariance_matrix,
                interleaved=False,
            )
            posterior = GPyTorchPosterior(distribution=mtmvn)
        if hasattr(self, "outcome_transform"):
            posterior = self.outcome_transform.untransform_posterior(posterior, X=X)
        if posterior_transform is not None:
            return posterior_transform(posterior)
        return posterior

    def subset_output(self, idcs: list[int]) -> MultiTaskGPyTorchModel:
        r"""Returns a new model that only outputs a subset of the outputs.

        Args:
            idcs: A list of output indices, corresponding to the outputs to keep.

        Returns:
            A new model that only outputs the requested outputs.
        """
        raise UnsupportedError(
            "Subsetting outputs is not supported by `MultiTaskGPyTorchModel`."
        )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
The base class for sampler modules to be used with MC-evaluated acquisition functions.
"""

from __future__ import annotations

from abc import ABC, abstractmethod

import torch
from botorch.exceptions.errors import InputDataError
from botorch.posteriors import Posterior
from torch import Tensor
from torch.nn import Module


KWARGS_DEPRECATED_MSG = (
    "The {} argument of `MCSampler`s has been deprecated and will raise an "
    "error in a future version."
)
KWARG_ERR_MSG = (
    "`MCSampler`s no longer support the `{}` argument. "
    "Consider using `{}` for similar functionality."
)


class MCSampler(Module, ABC):
    r"""Abstract base class for Samplers.

    Subclasses must implement the `forward` method.

    Example:
        This method is usually not called directly, but via the sampler's
        `__call__` method:
        >>> posterior = model.posterior(test_X)
        >>> samples = sampler(posterior)
    """

    def __init__(
        self,
        sample_shape: torch.Size,
        seed: int | None = None,
    ) -> None:
        r"""Abstract base class for samplers.

        Args:
            sample_shape: The `sample_shape` of the samples to generate. The full shape
                of the samples is given by `posterior._extended_shape(sample_shape)`.
            seed: An optional seed to use for sampling.
        """
        super().__init__()
        if not isinstance(sample_shape, torch.Size):
            raise InputDataError(
                "Expected `sample_shape` to be a `torch.Size` object, "
                f"got {sample_shape}."
            )
        self.sample_shape = sample_shape
        self.seed = seed if seed is not None else torch.randint(0, 1000000, (1,)).item()
        self.register_buffer("base_samples", None)

    @abstractmethod
    def forward(self, posterior: Posterior) -> Tensor:
        r"""Draws MC samples from the posterior.

        Args:
            posterior: The posterior to sample from.

        Returns:
            The samples drawn from the posterior.
        """
        pass  # pragma no cover

    def _get_batch_range(self, posterior: Posterior) -> tuple[int, int]:
        r"""Get the t-batch range of the posterior with an optional override.

        In rare cases, e.g., in `qMultiStepLookahead`, we may want to override the
        `batch_range` of the posterior. If this behavior is desired, one can set
        `batch_range_override` attribute on the samplers.

        Args:
            posterior: The posterior to sample from.

        Returns:
            The t-batch range to use for collapsing the base samples.
        """
        if hasattr(self, "batch_range_override"):
            return self.batch_range_override
        return posterior.batch_range

    def _get_collapsed_shape(self, posterior: Posterior) -> torch.Size:
        r"""Get the shape of the base samples with the t-batches collapsed.

        Args:
            posterior: The posterior to sample from.

        Returns:
            The collapsed shape of the base samples expected by the posterior. The
            t-batch dimensions of the base samples are collapsed to size 1. This is
            useful to prevent sampling variance across t-batches.
        """
        base_sample_shape = posterior.base_sample_shape
        batch_start, batch_end = self._get_batch_range(posterior)
        base_sample_shape = (
            base_sample_shape[:batch_start]
            + torch.Size([1 for _ in base_sample_shape[batch_start:batch_end]])
            + base_sample_shape[batch_end:]
        )
        return self.sample_shape + base_sample_shape

    def _get_extended_base_sample_shape(self, posterior: Posterior) -> torch.Size:
        r"""Get the shape of the base samples expected by the posterior.

        Args:
            posterior: The posterior to sample from.

        Returns:
            The extended shape of the base samples expected by the posterior.
        """
        return self.sample_shape + posterior.base_sample_shape

    def _update_base_samples(
        self, posterior: Posterior, base_sampler: MCSampler
    ) -> None:
        r"""Update the sampler to use the original base samples for X_baseline.

        This is used in CachedCholeskyAcquisitionFunctions to ensure consistency.

        Args:
            posterior: The posterior for which the base samples are constructed.
            base_sampler: The base sampler to retrieve the base samples from.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} does not implement `_update_base_samples`."
        )

    def _instance_check(self, base_sampler):
        r"""Check that `base_sampler` is an instance of `self.__class__`."""
        if not isinstance(base_sampler, self.__class__):
            raise RuntimeError(
                "Expected `base_sampler` to be an instance of "
                f"{self.__class__.__name__}. Got {base_sampler}."
            )
